from juriscraper.OpinionSite import OpinionSite
from datetime import datetime


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = "http://www.utcourts.gov/opinions/supopin/index.htm"
        self.court_id = self.__module__

    def _get_case_names(self):
        return [
            name
            for name in self.html.xpath(
                '/html/body//div[@id="content"]//p[a[@class="bodylink"]]/a/text()'
            )
        ]

    def _get_download_urls(self):
        return [
            t
            for t in self.html.xpath(
                '/html/body//div[@id="content"]//p[a[@class="bodylink"]]/a/@href'
            )
        ]

    def _get_docket_numbers(self):
        docket_numbers = []
        for text in self.html.xpath(
            '/html/body//div[@id="content"]//p[a[@class="bodylink"]]/text()'
        ):
            try:
                parts = text.strip().split(", ")
                docket_numbers.append(parts[1])
            except IndexError:
                # Happens in whitespace-only text nodes.
                continue
        return docket_numbers

    def _get_case_dates(self):
        dates = []
        for text in self.html.xpath(
            '/html/body//div[@id="content"]//p[a[@class="bodylink"]]/text()'
        ):
            parts = text.strip().split(", ")
            try:
                caseDate = parts[-3] + ", " + parts[-2]
                dates.append(datetime.strptime(caseDate, "Filed %B %d, %Y"))
            except IndexError:
                # Happens in whitespace-only text nodes.
                continue
        return dates

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)

    def _get_neutral_citations(self):
        neutral_citations = []
        for text in self.html.xpath(
            '/html/body//div[@id="content"]//p[a[@class="bodylink"]]/text()'
        ):
            try:
                parts = text.strip().split(", ")
                if parts[-1]:
                    neutral_citations.append(parts[-1])
            except IndexError:
                # Happens in whitespace-only text nodes.
                continue
        return neutral_citations



# -*- coding: utf-8 -*-
# Scraper for the United States Tax Court
# CourtID: tax
# Court Short Name: Tax Ct.
# Neutral Citation Format (Tax Court opinions): 138 T.C. No. 1 (2012)
# Neutral Citation Format (Memorandum opinions): T.C. Memo 2012-1
# Neutral Citation Format (Summary opinions: T.C. Summary Opinion 2012-1

import re
from datetime import date, datetime, timedelta
from dateutil.rrule import WEEKLY, rrule

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteWebDriven import OpinionSiteWebDriven
from juriscraper.lib.html_utils import fix_links_but_keep_anchors
from juriscraper.lib.models import citation_types


class Site(OpinionSiteWebDriven):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.uses_selenium = True
        self.url = "https://www.ustaxcourt.gov/UstcInOp/OpinionSearch.aspx"
        self.base_path = (
            '//tr[@class="ResultsOddRow" or ' '@class="ResultsEvenRow"]'
        )
        self.case_date = date.today()
        self.backwards_days = 14
        self.back_scrape_iterable = [
            i.date()
            for i in rrule(
                WEEKLY, dtstart=date(1995, 9, 25), until=date(2018, 11, 13),
            )
        ]

        self.court_id = self.__module__

    def _download(self, request_dict={}):
        """Uses Selenium because doing it with requests is a pain."""
        if self.test_mode_enabled():
            return super(Site, self)._download(request_dict=request_dict)

        logger.info("Now downloading case page at: %s" % self.url)
        self.initiate_webdriven_session()

        # Set the start and end dates
        start_date_id = "ctl00_Content_dpDateSearch_dateInput"
        start_date_input = self.webdriver.find_element_by_id(start_date_id)
        start_date_input.send_keys(
            (self.case_date - timedelta(days=self.backwards_days)).strftime(
                "%-m/%-d/%Y"
            )
        )

        end_date_id = "ctl00_Content_dpDateSearchTo_dateInput"
        end_date_input = self.webdriver.find_element_by_id(end_date_id)
        end_date_input.send_keys(self.case_date.strftime("%-m/%-d/%Y"))
        # self.take_screenshot()

        # Check ordering by case date (this orders by case date, *ascending*)
        ordering_id = "Content_rdoCaseName_1"
        self.webdriver.find_element_by_id(ordering_id).click()

        # Submit
        self.webdriver.find_element_by_id("Content_btnSearch").click()

        # Do not proceed until the results show up.
        self.wait_for_id("Content_ddlResultsPerPage")
        # self.take_screenshot()

        text = self._clean_text(self.webdriver.page_source)
        self.webdriver.quit()
        html_tree = self._make_html_tree(text)
        html_tree.rewrite_links(fix_links_but_keep_anchors, base_href=self.url)
        return html_tree

    def _get_download_urls(self):
        # URLs here take two forms. For precedential cases, they're simple, and
        # look like:
        #   https://www.ustaxcourt.gov/UstcInOp/OpinionViewer.aspx?ID=11815
        # But for non-precedential cases, they have an annoying JS thing, and
        # links look like:
        #   https://www.ustaxcourt.gov/UstcInOp/OpinionSearch.aspx#11813
        # Note the hash instead of the ID.
        #
        # This is annoying, but we just have to swap out the ending and it
        # should be fine.
        hrefs = []
        path = self.base_path + "//@href"
        for href in self.html.xpath(path):
            if "?ID" in href:
                hrefs.append(href)
            else:
                hrefs.append(
                    href.replace(
                        "OpinionSearch.aspx#", "OpinionViewer.aspx?ID="
                    )
                )
        return hrefs

    def _get_case_names(self):
        case_names = []
        path = self.base_path + "//td[1]"
        for td in self.html.xpath(path):
            case_names.append(td.text_content().strip() + " v. Commissioner")
        return case_names

    def _get_case_name_shorts(self):
        # The normal values are particularly bad, usually just returning
        # "Commissioner" for all cases. Just nuke these values.
        return [""] * len(self.case_names)

    def _get_precedential_statuses(self):
        statuses = []
        path = self.base_path + "//td[2]"
        for td in self.html.xpath(path):
            status = td.text_content().strip().lower()
            if "opinion" in status:
                statuses.append("Published")
            elif "memorandum" in status:
                statuses.append("Unpublished")
            elif "summary" in status:
                statuses.append("Unpublished")
            else:
                statuses.append("Unknown")
        return statuses

    def _get_case_dates(self):
        dates = []
        path = self.base_path + "//td[3]"
        for td in self.html.xpath(path):
            date_string = td.text_content().strip()
            dates.append(datetime.strptime(date_string, "%m/%d/%Y").date())
        return dates

    def _download_backwards(self, d):
        self.backwards_days = 7
        self.case_date = d
        logger.info(
            "Running backscraper with date range: %s to %s",
            self.case_date - timedelta(days=self.backwards_days),
            self.case_date,
        )
        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200

    def extract_from_text(self, scraped_text):
        """Can we extract the citation and related information

        :param scraped_text: The content of the docuemnt downloaded
        :return: dictionary of citations, reporter, volume and page
        """
        metadata = {
            "Citation": {"type": citation_types["SPECIALTY"]},
            "Docket": {"docket_number": ""},
            "OpinionCluster": {"precedential_status": ""},
        }

        tax_court_reports_regex = re.compile(
            r"""
            ([0-9]{1,4})\s{1,}                           # (volume)
            UNITED\ STATES\ TAX\ COURT\ REPORTS?\s{1,}   # (reporter)
            \((\d{1,4})\)                                # (page)
            """,
            re.VERBOSE | re.IGNORECASE,
        )

        tax_court_alt_regex = re.compile(
            r"""
            ((T\.\ ?C\.\s((Memo\.?)|(Summm?ary\ Opinion))\s{1,} # T.C. Memo or Summary Opinion (reporter)
            ([0-9]{4})                                          # Four digit year (volume)
            .                                                   # hyphen, em-dash etc.
            ([0-9]{1,3})\b)                                     # 1-3 digit number in order of publication (page)
            |                                                   # or
            ([0-9]{1,4})\s{1,}                                  # (volume)
            (T\.\ ?C\.\ No\.)(?:\s{1,})?                        # T.C. No. (reporter)
            (\d{1,4}))                                          # (page)
            """,
            re.VERBOSE | re.IGNORECASE,
        )

        match = re.search(tax_court_reports_regex, scraped_text)

        if match:
            metadata["Citation"]["volume"] = match.group(1)
            metadata["Citation"]["page"] = match.group(2)
            metadata["Citation"]["reporter"] = "T.C."
            metadata["OpinionCluster"]["precedential_status"] = "Published"
        else:
            match = re.search(tax_court_alt_regex, scraped_text)
            if match:
                if "No." in match.group():
                    metadata["Citation"]["reporter"] = "T.C. No."
                    metadata["Citation"]["volume"] = match.group(8)
                    metadata["Citation"]["page"] = match.group(10)
                    metadata["OpinionCluster"][
                        "precedential_status"
                    ] = "Published"
                else:
                    if "Memo" in match.group():
                        metadata["Citation"]["reporter"] = "T.C. Memo."
                    elif "Summ" in match.group():
                        metadata["Citation"][
                            "reporter"
                        ] = "T.C. Summary Opinion"
                    metadata["Citation"]["volume"] = match.group(6)
                    metadata["Citation"]["page"] = match.group(7)
                    metadata["OpinionCluster"][
                        "precedential_status"
                    ] = "Unpublished"

        metadata["Docket"]["docket_number"] = self.get_tax_docket_numbers(
            scraped_text
        )
        return metadata

    def get_tax_docket_numbers(self, opinion_text):
        """Parse opinion plain text for docket numbers.

        First we idenitify where the docket numbers are in the document.
        This is normally at the start of the document but can often follow
         a lengthy case details section.

        :param opinion_text: is the opinions plain_text
        :return docket_string: as string of docket numbers Ex. (18710-94, 12321-95)
        """
        opinion_text = self.remove_en_em_dash(opinion_text)
        parsed_text = ""
        docket_no_re = r"Docket.? Nos?.? .*[0-9]{3,5}"
        matches = re.finditer(docket_no_re, opinion_text)

        for matchNum, match in enumerate(matches, start=1):
            parsed_text = opinion_text[match.start() :]
            break

        matches2 = re.finditer(
            r"[0-9]{3,5}(-|–)[\w]{2,4}([A-Z])?((\.)| [A-Z]\.)", parsed_text
        )
        for m2, match2 in enumerate(matches2, start=0):
            parsed_text = parsed_text[: match2.end()]
            break

        docket_end_re = r"[0-9]{3,5}(-|–)[\w]{2,4}([A-Z])?((\,|\.)| [A-Z]\.)"

        matches = re.finditer(docket_end_re, parsed_text, re.MULTILINE)
        hits = []
        for matchNum, match in enumerate(matches, start=1):
            hits.append(match.group())
        docket_string = ", ".join(hits).replace(",,", ",").replace(".", "")
        return docket_string.strip()

    def remove_en_em_dash(self, opinion_text):
        opinion_text = re.sub(u"–", "-", opinion_text)
        opinion_text = re.sub(u"—", "-", opinion_text)
        opinion_text = re.sub(u"–", "-", opinion_text)
        return opinion_text



# -*- coding: utf-8 -*-

from juriscraper.OpinionSite import OpinionSite
from juriscraper.WebDriven import WebDriven


class OpinionSiteWebDriven(OpinionSite, WebDriven):
    def __init__(self, *args, **kwargs):
        super(OpinionSiteWebDriven, self).__init__(*args, **kwargs)
        WebDriven.__init__(self, args, kwargs)



"""Scraper for D.C. Circuit of Appeals
CourtID: cadc
Court Short Name: cadc
Author: Andrei Chelaru
Reviewer: mlr
Date created: 18 July 2014
"""

from datetime import datetime, date

from juriscraper.OralArgumentSite import OralArgumentSite


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        d = date.today()
        # d = date(month=5, day=1, year=2014)
        self.url = "http://www.cadc.uscourts.gov/recordings/recordings.nsf/DocsByRDate?OpenView&count=100&SKey={yearmo}".format(
            yearmo=d.strftime("%Y%m")
        )
        self.back_scrape_iterable = [
            "%s%02d" % (year, month)
            for year in range(2007, d.year + 1)
            for month in range(1, 13)
        ]

    def _download(self, **kwargs):
        # The certificate on their site has expired.
        return super(Site, self)._download(request_dict={"verify": False})

    def _get_download_urls(self):
        path = "id('ViewBody')//div[contains(concat(' ',@class,' '),' row-entry')]//@href"
        return list(self.html.xpath(path))

    def _get_case_names(self):
        path = "id('ViewBody')//*[contains(concat(' ',@class,' '),' column-two')]/div[1]/text()"
        return list(self.html.xpath(path))

    def _get_case_dates(self):
        path = "id('ViewBody')//date/text()"
        return map(self._return_case_date, self.html.xpath(path))

    @staticmethod
    def _return_case_date(e):
        e = "".join(e.split())
        return datetime.strptime(e, "%m/%d/%Y").date()

    def _get_docket_numbers(self):
        path = "id('ViewBody')//*[contains(concat(' ',@class,' '),' row-entry')]//a//text()"
        return list(self.html.xpath(path))

    def _get_judges(self):
        path = '//div[span[contains(., "Judges")]]/text()'
        return [" ".join(s.split()) for s in self.html.xpath(path)]

    def _download_backwards(self, yearmo):
        self.url = "http://www.cadc.uscourts.gov/recordings/recordings.nsf/DocsByRDate?OpenView&count=100&SKey={yearmo}".format(
            yearmo=yearmo,
        )
        self.html = self._download()



from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    """This court provides a json endpoint!
    How fun. Their single json endpoint dumps
    all of the cases into the following groupings:
        - Supreme
        - Published_Appellate
        - Unpublished_Appellate
        - Published_tax
        - Unpublished_Tax
        - Unpublished_Trial

    Human web interface: http://www.judiciary.state.nj.us/attorneys/opinions.html#Supreme
    """

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.base_url = "http://www.judiciary.state.nj.us/attorneys/assets"
        self.servername = "http://www.judiciary.state.nj.us"
        self.url = "%s/js/objects/opinions/op2017.json" % self.base_url
        self.case_types = ["Supreme"]

    def tweak_response_object(self):
        """Court is irresponsibly not returning valid json content
        type header, so we force it here. As a result, the self.html
        object will be a dict of the parsed json data.
        """
        self.request["response"].headers["content-type"] = "application/json"

    def _get_download_urls(self):
        urls = []
        for type in self.case_types:
            for opinion in self.html[type]:
                # They don't have uniform keys across all data, which is... odd
                if "DocumentURL" in opinion:
                    url = "%s%s" % (self.servername, opinion["DocumentURL"])
                else:
                    url = self.get_absolute_opinion_path(
                        opinion["Document"], type
                    )
                urls.append(url)
        return urls

    def _get_case_names(self):
        names = []
        for type in self.case_types:
            names.extend([case["Title"] for case in self.html[type]])
        return names

    def _get_case_dates(self):
        dates = []
        for type in self.case_types:
            dates.extend(
                [
                    convert_date_string(case["PublishDate"])
                    for case in self.html[type]
                ]
            )
        return dates

    def _get_precedential_statuses(self):
        statuses = []
        for type in self.case_types:
            status = (
                "Unpublished"
                if type.startswith("Unpublished")
                else "Published"
            )
            statuses.extend([status] * len(self.html[type]))
        return statuses

    def _get_docket_numbers(self):
        dockets = []
        for type in self.case_types:
            dockets.extend([case["OpinionID"] for case in self.html[type]])
        return dockets

    def get_absolute_opinion_path(self, suffix, type):
        """Determine the absolute path given the file suffix in the
        json object and the opinion type.  This is necessary because
        the course does not return standardized data objects.
        """
        type_parts = type.lower().split("_")
        type_parts_length = len(type_parts)
        if type_parts_length == 1:
            status = False
            type = type_parts[0].lower()
        elif type_parts_length == 2:
            status = type_parts[0].lower()
            type = type_parts[1].lower()
        else:
            raise InsanityException(
                'Unrecognized type "%s", this should never ' "happen" % type
            )
        if not suffix.startswith(type):
            if status:
                suffix = "%s/%s/%s" % (type, status, suffix)
            else:
                suffix = "%s/%s" % (type, suffix)
        return "%s/opinions/%s" % (self.base_url, suffix)



"""Scraper for Maryland Supreme Court Oral Argument Audio

This scraper has an interesting history. It was briefly running on the live
site, but we realized shortly after starting it that the scraper was
downloading video, not audio!

Seeing that we weren't ready for video, we disabled this scraper and deleted
any traces of it on the server.

One interesting lesson though was that the OA system didn't crumble or budge
when this was running. The video converted to mp3 just fine (each item took a
few hours) and we began hosting it like nothing was different. Go figure.

Your humble editor,

Mike


CourtID: md
Court Short Name: Md.
"""

from juriscraper.OralArgumentSiteLinear import OralArgumentSiteLinear


class Site(OralArgumentSiteLinear):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "http://www.courts.state.md.us/coappeals/webcasts/webcastarchive.html"

    def _process_html(self):
        # Find rows that contain valid, non-"Bar Admissions", link
        path = (
            "//tr[.//td[2]//a/@href][not(contains(.//@href, 'baradmission'))]"
        )
        rows = self.html.xpath(path)

        for row in rows:
            cell_two = row.xpath("./td[2]")[0]
            self.cases.append(
                {
                    "date": row.xpath("./td[1]")[0].text_content(),
                    "name": row.xpath("./td[3]/*[self::b or self::strong]")[
                        0
                    ].text_content(),  # sometimes they use <b> tag, other times <strong> tag
                    "docket": cell_two.text_content(),
                    "url": cell_two.xpath(".//a/@href")[0],
                }
            )



"""Scraper for the California Attorney General
CourtID: nyag
Court Short Name: New York Attorney General
"""

import datetime

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.year = datetime.date.today().year
        self.base_url = "https://ag.ny.gov/appeals-and-opinions/numerical-index?field_opinion_year_value=%d"
        self.url = self.base_url % self.year
        self.back_scrape_iterable = range(1995, self.year + 1)
        self.row_path = False
        self.cell_path = False
        self.set_paths()

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        if self.test_mode_enabled():
            # Make sure the year-table you want to test is first in example file
            self.year = int(html.xpath("//table[1]/caption")[0].text_content())
            self.set_paths()
        return html

    def _get_case_dates(self):
        """All we have are years, so estimate middle most day of year"""
        today = datetime.date.today()
        middle_of_year = convert_date_string("July 2, %d" % self.year)
        if self.year == today.year:
            # Not a backscraper, assume cases were filed on day scraped.
            return [today] * len(self.html.xpath(self.row_path))
        else:
            return [middle_of_year] * len(self.html.xpath(self.row_path))

    def _get_case_names(self):
        """No case names available"""
        return ["Untitled New York Attorney General Opinion"] * len(
            self.case_dates
        )

    def _get_download_urls(self):
        path = "%s//a/@href" % (self.cell_path % 4)
        return [href for href in self.html.xpath(path)]

    def _get_docket_numbers(self):
        return [
            cell.text_content().strip()
            for cell in self.html.xpath(self.cell_path % 1)
        ]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_dates)

    def _get_summaries(self):
        """Use Abstract column value"""
        return [
            cell.text_content().strip()
            for cell in self.html.xpath(self.cell_path % 2)
        ]

    def _get_date_filed_is_approximate(self):
        return [True] * len(self.case_dates)

    def _download_backwards(self, year):
        self.year = year
        self.url = self.base_url % self.year
        self.set_paths()
        self.html = self._download()

    def set_paths(self):
        self.row_path = '//table[contains(caption, "%d")]/tbody/tr' % self.year
        self.cell_path = self.row_path + "/td[%d]"



"""Scraper for Board of Veterans' Appeals
CourtID: bva
Court Short Name: BVA
Author: Jon Andersen
Reviewer: mlr
Type: Nonprecedential
History:
    2014-09-09: Created by Jon Andersen
    2016-05-14: Updated by arderyp, moved logic from _get_case_dates() to
    standard download() override method.  Parsing information text from
    unlinked text field, due to recent link text typos
"""

import re
from datetime import datetime

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        # Cases before 1997 do not have a docket number to parse.
        url_query = "&DB=".join(
            [str(n) for n in range(datetime.today().year, 1997 - 1, -1)]
        )
        self.url = (
            "http://www.index.va.gov/search/va/bva_search.jsp?RPP=50&RS=1&DB=%s"
            % url_query
        )
        self.pager_stop = False
        self.back_scrape_iterable = self.pager(50)
        self.cases = []

    def pager(self, incr):
        startat = 1 + incr
        while not self.pager_stop:
            yield startat
            startat += incr

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self._extract_case_data_from_html(html)
        return html

    def _extract_case_data_from_html(self, html):
        """Build list of data dictionaries, one dictionary per case."""
        regex = re.compile(
            r"^Citation Nr: (.*) Decision Date: (.*) Archive Date: (.*) DOCKET NO. ([-0-9 ]+)"
        )

        for result in html.xpath('//div[@id="results-area"]/div/a'):
            text = result.text_content().strip()
            try:
                (citation, date, docket) = regex.match(text).group(1, 2, 4)
            except:
                raise Exception(
                    "regex failure in _extract_case_data_from_html method of bva scraper"
                )

            # There is a history to this, but the long story short is that we
            # are using the docket number in the name field intentionally.
            self.cases.append(
                {
                    "name": docket,
                    "url": result.xpath(".//@href")[0],
                    "date": convert_date_string(date),
                    "status": "Unpublished",
                    "docket": docket,
                    "citation": citation.split()[0],
                }
            )

    def _get_case_dates(self):
        return [case["date"] for case in self.cases]

    def _get_download_urls(self):
        return [case["url"] for case in self.cases]

    def _get_case_names(self):
        return [case["name"] for case in self.cases]

    def _get_docket_numbers(self):
        return [case["docket"] for case in self.cases]

    def _get_precedential_statuses(self):
        return [case["status"] for case in self.cases]

    def _download_backwards(self, startat):
        base_url = (
            "http://www.index.va.gov/search/va/bva_search.jsp?RPP=50&RS=%d"
            % (startat,)
        )
        date_range = range(datetime.today().year, 1997 - 1, -1)
        self.url = (
            base_url + "&DB=" + "&DB=".join([str(n) for n in date_range])
        )
        self.html = self._download()



# Scraper for the United States Court of International Trade
# CourtID: cit
# Court Short Name: Ct. Int'l Trade
# Neutral Citation Format: Ct. Int'l Trade No. 12-1

from juriscraper.OpinionSite import OpinionSite
import time
from datetime import date
from lxml import html


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = "http://www.cit.uscourts.gov/SlipOpinions/index.html"
        self.court_id = self.__module__
        self.base = '//tr[../tr/th[contains(., "Caption")]]'

    def _get_download_urls(self):
        return [
            t
            for t in self.html.xpath(
                "{base}/td[1]/a/@href".format(base=self.base)
            )
        ]

    def _get_neutral_citations(self):
        neutral_citations = []
        for t in self.html.xpath(
            "{base}/td[1]/a/text()".format(base=self.base)
        ):
            year, item_number = t.split("-")
            neutral_citations.append(
                "20{year} CIT {number}".format(year=year, number=item_number)
            )
        return neutral_citations

    def _get_case_names(self):
        # Exclude confidential rows by ensuring there is a sibling row that
        # contains an anchor (which confidential cases do not)
        case_names = []
        for e in self.html.xpath(
            "{base}/td[2][../td/a]".format(base=self.base)
        ):
            text_nodes = e.xpath(".//text()")
            case_names.append(text_nodes[0])
        return case_names

    def _get_precedential_statuses(self):
        statuses = []
        for e in self.html.xpath(
            "{base}/td[2][../td/a]".format(base=self.base)
        ):
            s = (
                html.tostring(e, method="text", encoding="unicode")
                .lower()
                .strip()
            )
            if "errata" in s:
                statuses.append("Errata")
            else:
                statuses.append("Published")
        return statuses

    def _get_case_dates(self):
        # This does not capture the release dates for the errata documents.
        # The errata release date is listed in column 2. This will use the
        # original release date instead.
        dates = []
        date_formats = ["%m/%d/%Y", "%m/%d/%y"]
        for date_string in self.html.xpath(
            "{base}/td[3][../td/a]//text()".format(base=self.base)
        ):
            for date_format in date_formats:
                try:
                    d = date.fromtimestamp(
                        time.mktime(
                            time.strptime(date_string.strip(), date_format)
                        )
                    )
                    dates.append(d)
                    break
                except ValueError:
                    # Try the next format
                    continue
        return dates

    def _get_docket_numbers(self):
        docket_numbers = []
        for e in self.html.xpath(
            "{base}/td[4][../td/a]".format(base=self.base)
        ):
            docket_numbers.append(
                html.tostring(e, method="text", encoding="unicode").strip()
            )
        return docket_numbers

    def _get_judges(self):
        judges = []
        for e in self.html.xpath(
            "{base}/td[5][../td/a]".format(base=self.base)
        ):
            s = html.tostring(e, method="text", encoding="unicode")
            judges.append(s)
        return judges

    def _get_nature_of_suit(self):
        return [
            t
            for t in self.html.xpath(
                "{base}/td[6][../td/a]/text()".format(base=self.base)
            )
        ]



"""Scraper for the Rhode Island Supreme Court
CourtID: ri
Court Short Name: R.I.
Court Contact: helpdesk@courts.ri.gov, MFerris@courts.ri.gov (Ferris, Mirella), webmaster@courts.ri.gov
    https://www.courts.ri.gov/PDF/TelephoneDirectory.pdf
Author: Brian W. Carver
Date created: 2013-08-10
"""

import re
from datetime import datetime
from juriscraper.OpinionSiteLinear import OpinionSiteLinear
from juriscraper.lib.exceptions import InsanityException


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = self.build_url(
            "https://www.courts.ri.gov/Courts/SupremeCourt/Pages/Opinions/Opinions"
        )
        self.previous_date = None
        self.include_summary = True
        self.status = "Published"

        # HTTPS certificate is bad, but hopefully they'll fix it and we can remove the line below
        self.disable_certificate_verification()

    def build_url(self, base_url):
        # This court hears things from mid-September to end of June. This
        # defines the "term" for that year, which triggers the website updates.
        today = datetime.today()
        this_year = today.year
        term_end = datetime(this_year, 9, 15)
        year = this_year if today >= term_end else this_year - 1
        return "%s%d-%d.aspx" % (base_url, year, year + 1)

    def _process_html(self):
        # case information spans over 3 rows, so must process 3 at a time:
        #   <tr> - contains case name, docket number, date and pdf link
        #   <tr> - contains case summary
        #   <tr> - contains a one-pixel gif spacer
        table = "//table[@id = 'onetidDoclibViewTbl0']/tr[position() > 1]"
        rows = list(self.html.xpath(table))
        row_triplets = list(zip(rows, rows[1:]))[::3]

        for tr1, tr2 in row_triplets:
            case = self.extract_case_from_rows(tr1, tr2)
            self.previous_date = case["date"]
            self.cases.append(case)

    def extract_case_from_rows(self, row1, row2):
        docket = row1.xpath("./td/a/text()")[0]
        docket = ", ".join([d.strip() for d in docket.split(",")])
        url = row1.xpath("./td/a/@href")[0]
        text = row1.xpath("./td[1]/text()")[0]
        text_to_parse = [text]

        if self.include_summary:
            summary_lines = row2.xpath("./td/div/text()")
            summary = "\n".join(summary_lines)
            joined_text = "\n".join([text, summary_lines[0]])
            text_to_parse.append(joined_text)
        else:
            summary = False

        return {
            "url": url,
            "docket": docket,
            "date": self.parse_date_from_text(text_to_parse),
            "name": self.parse_name_from_text(text_to_parse),
            "summary": summary,
        }

    def parse_date_from_text(self, text_list):
        regex = "(.*?)(\((\w+\s+\d+\,\s+\d+)\))(.*?)"
        for text in text_list:
            date_match = re.match(regex, text)
            if date_match:
                return date_match.group(3)

        # Fall back on previous case's date
        if self.previous_date:
            return self.previous_date

        raise InsanityException(
            "Could not parse date from string, and no "
            'previous date to fall back on: "%s"' % text_list
        )

    @staticmethod
    def parse_name_from_text(text_list):
        regexes = [
            # Expected format
            "(.*?)(,?\sNos?\.)(.*?)",
            # Clerk typo, forgot "No."/"Nos." substring
            "(.*?)(,?\s\d+-\d+(,|\s))(.*?)",
            # Same as above, and there's an unconventional docket number
            # like 'SU-14-324' instead of '14-324'. See ri_p_example_4.html
            "(.*?)(,?\s(?:\w+-)?\d+-\d+(,|\s))(.*?)",
        ]

        for regex in regexes:
            for text in text_list:
                name_match = re.match(regex, text)
                if name_match:
                    return name_match.group(1)

        # "No."/"Nos." and docket missing, fall back on whatever's before first
        # semi-colon
        for text in text_list:
            if ";" in text:
                return text.split(";")[0]

        raise InsanityException(
            'Could not parse name from string: "%s"' % text_list
        )



# Scraper for California's First District Court of Appeal
# CourtID: calctapp_1st
# Court Short Name: Cal. Ct. App.

from juriscraper.opinions.united_states.state import cal


class Site(cal.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_code = "A"
        self.division = "1st App. Dist."
        self.url = self.build_url()

    def _get_divisions(self):
        return [self.division] * len(self.case_names)



"""
Scraper for the Tennessee Court of Appeals
CourtID: tennctapp
Court Short Name: Tenn. Ct. App.
"""
from juriscraper.opinions.united_states.state import tenn


class Site(tenn.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "http://www.tsc.state.tn.us/courts/court-appeals/opinions"
        self.back_scrape_iterable = range(0, 987)



"""Scraper for the D.C. Court of Appeals
CourtID: dc
Court Short Name: D.C.
Author: V. David Zvenyach
Date created:2014-02-21
"""

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.dccourts.gov/court-of-appeals/opinions-memorandum-of-judgments"
        qualifier_no_opinions = (
            'not(contains(td[2]/span/text(), "NO OPINIONS"))'
        )
        qualifier_has_pdf_link = 'contains(.//td[1]/a/@href, ".pdf")'
        self.base_path = "//table//tr[%s and %s]" % (
            qualifier_no_opinions,
            qualifier_has_pdf_link,
        )

    def _get_docket_numbers(self):
        path = "%s/td[1]/a" % self.base_path
        return [cell.text_content().strip() for cell in self.html.xpath(path)]

    def _get_download_urls(self):
        path = "%s/td[1]/a/@href" % self.base_path
        return [href for href in self.html.xpath(path)]

    def _get_case_names(self):
        path = "%s/td[2]" % self.base_path
        return [cell.text_content() for cell in self.html.xpath(path)]

    def _get_case_dates(self):
        path = "%s/td[3]" % self.base_path
        return [
            convert_date_string(cell.text_content())
            for cell in self.html.xpath(path)
        ]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)



"""Scraper for the Maryland Attorney General
CourtID: ag
Court Short Name: Maryland Attorney General
"""

import datetime
from time import sleep
from lxml import html
from selenium.common.exceptions import NoSuchElementException

from juriscraper.OpinionSiteWebDriven import OpinionSiteWebDriven
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSiteWebDriven):
    """This scraper is strange. The site it temperamental, and the javascript
    seems to load successfully on some runs, but not on others. The dates are
    also estimated, and the names are actually semi-long summaries. Furthermore,
    the site's source is unmanageable, which has prevented us from being able to
    create legitimate test/example files for coverage. We have a single example
    file that's an empty document skeleton to prevent the test mechanism from
    complaining. But it isn't a test providing real coverage.

    We are doing the best we can with a bad site.
    """

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.year = datetime.date.today().year
        self.domain = "http://www.marylandattorneygeneral.gov"
        self.url = "%s/Pages/Opinions/index.aspx" % self.domain
        self.back_scrape_iterable = range(1993, self.year + 1)
        self.parent_path_base = '//tbody/tr/td[contains(./text(), "%d")]'
        self.parent_path = self.parent_path_base % self.year
        self.cell_path = '//tbody[@isloaded="true"]/tr/td[%d]'
        self.next_path = '//a[@title="Next"]'

    def _download(self, request_dict={}):
        if self.test_mode_enabled():
            return [super(Site, self)._download(request_dict)]
        trees = self.get_dynamic_html_trees()
        if not len(trees):
            # No opinions for current year on page, SO no
            # js to load.  Return regular page html and
            # extract 0 cases because nothing there
            return [super(Site, self)._download(request_dict)]
        return trees

    def get_dynamic_html_trees(self):
        # Initialize webdriver
        self.initiate_webdriven_session()

        # Find and activate the opinion drop-down for year
        try:
            date_anchor = self.webdriver.find_element_by_xpath(
                "%s/a" % self.parent_path
            )
        except NoSuchElementException:
            # Year has no opinions drop-down on page
            return []
        date_anchor.click()
        trees = [self.get_tree_from_driver_dom()]

        # Handle pagination if more than 30 results for year
        while True:
            try:
                next_anchor = self.webdriver.find_element_by_xpath(
                    self.next_path
                )
            except NoSuchElementException:
                # Less than 30 results
                break
            next_anchor.click()
            trees.append(self.get_tree_from_driver_dom())
        return trees

    def get_tree_from_driver_dom(self):
        # Wait for js to load and dom html to update
        # Seems stupid, but necessary, and easier
        # thank loading lots of selenium dependencies
        # and using complex WebDriverWait with callbacks
        # for attribute to appear, which don't even
        # seem to work consistently with the site's
        # finicky responses.
        sleep(3)
        source = self.webdriver.execute_script(
            "return document.getElementsByTagName('html')[0].innerHTML"
        )
        tree = html.fromstring(source)
        tree.make_links_absolute(self.domain)
        return tree

    def _get_case_names(self):
        names = []
        path = self.cell_path % 3
        for tree in self.html:
            names.extend(
                [cell.text_content().strip() for cell in tree.xpath(path)]
            )
        return names

    def _get_download_urls(self):
        urls = []
        path = (self.cell_path % 4) + "/a/@href"
        for tree in self.html:
            urls.extend([href for href in tree.xpath(path)])
        return urls

    def _get_case_dates(self):
        today = datetime.date.today()
        count = len(self._get_case_names())
        middle_of_year = convert_date_string("July 2, %d" % self.year)
        if self.year == today.year:
            # Not a backscraper, assume cases were filed on day scraped.
            return [today] * count
        else:
            # All we have is the year, so estimate the middle most day
            return [middle_of_year] * count

    def _get_docket_numbers(self):
        dockets = []
        path = self.cell_path % 1
        for tree in self.html:
            for cell in tree.xpath(path):
                dockets.append(cell.text_content().replace("Unpublished", ""))
        return dockets

    def _get_precedential_statuses(self):
        statuses = []
        path = self.cell_path % 1
        for tree in self.html:
            for cell in tree.xpath(path):
                if "Unpublished" in cell.text_content():
                    statuses.append("Unpublished")
                else:
                    statuses.append("Published")
        return statuses

    def _get_date_filed_is_approximate(self):
        return ["True"] * len(self.case_names)

    def _download_backwards(self, year):
        """Iterate over drop down for each year on the page"""
        self.year = year
        self.parent_path = self.parent_path_base % year
        self.html = self._download()



# -*- coding: utf-8 -*-
import datetime
from six.moves.urllib import urlparse

import sys
from dateutil.rrule import rrule, DAILY
from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_name = (
            "United States District Court Eastern District of Louisiana"
        )
        today = datetime.date.today().strftime("%Y-%m-%d")

        self.base_url = "http://www.gpo.gov"
        self.query = 'collection:USCOURTS and courttype:District and courtname:"{court}" and publishdate:{date} and accode:USCOURTS'
        self.url_template = "http://www.gpo.gov/fdsys/search/uscourtsFwdSearch.action?forwadedQuery={}"
        self.url = self.url_template.format(
            # self.query.format(court=self.court_name, date="2015-10-26")
            self.query.format(court=self.court_name, date=today)
        )
        self.interval = 30
        self.back_scrape_iterable = [
            i.date()
            for i in rrule(
                DAILY,
                interval=self.interval,
                dtstart=datetime.date(2003, 2, 1),
                until=datetime.date(datetime.date.today().year + 1, 1, 1),
            )
        ]

        self.base_path = "//table[@class='search-results-item']"

    def _download(self, request_dict={}):
        """
        returns a list of html trees
        """

        html_tree = super(Site, self)._download(request_dict)
        tree_list = [
            html_tree,
        ]
        next_page = html_tree.xpath(
            "//div[@class='search-results-pagination']//a[contains(., 'Next')]/@href"
        )
        if next_page:
            self.url = urlparse.urljoin(self.base_url, next_page[0])
            tree_list.extend(self._download())
        return tree_list

    def _get_case_names(self):
        case_names = []
        for html_tree in self.html:
            for element in html_tree.xpath(self.base_path):
                case_name = "".join(
                    e.strip()
                    for e in element.xpath(
                        ".//a/span[@class='results-line1-title']//text()"
                    )
                )
                case_names.append(case_name.split(" - ")[-1])
        # print >>sys.stderr, case_names
        return case_names

    def _get_download_urls(self):
        download_urls = []
        for html_tree in self.html:
            for element in html_tree.xpath(self.base_path):
                download_urls.append(
                    "".join(
                        x
                        for x in element.xpath(
                            ".//a[span[@class='results-line1-pdf']]/@href"
                        )
                    )
                )
        return download_urls

    def _get_case_dates(self):
        case_dates = []
        for html_tree in self.html:
            for element in html_tree.xpath(self.base_path):
                case_date = "".join(
                    element.xpath(".//span[@class='results-line2']/text()")
                )
                case_date = case_date.split(",")
                case_date = "{}{}".format(case_date[-2], case_date[-1]).strip()
                # January 26, 2015.
                case_dates.append(
                    datetime.datetime.strptime(case_date.strip(), "%B %d %Y.")
                )
        return case_dates

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)

    def _get_docket_numbers(self):
        docket_numbers = []
        for html_tree in self.html:
            for element in html_tree.xpath(self.base_path):
                docket_number = "".join(
                    e.strip()
                    for e in element.xpath(
                        ".//a/span[@class='results-line1-title']//text()"
                    )
                )
                docket_numbers.append(docket_number.split(" - ")[0])
        return docket_numbers

    def _download_backwards(self, d):
        # range(2015-01-01,2015-02-01)
        date_range = "range({start_date},{end_date})".format(
            start_date=(d - datetime.timedelta(days=self.interval)).strftime(
                "%Y-%m-%d"
            ),
            end_date=d.strftime("%Y-%m-%d"),
        )
        self.url = self.url_template.format(
            self.query.format(court=self.court_name, date=date_range)
        )
        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200



from juriscraper.opinions.united_states.federal_appellate import scotus_slip


class Site(scotus_slip.Site):

    # Note that scotus_relating inherits from this class.
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.back_scrape_iterable = range(5, int(self.yy) + 1)
        self.precedential = "In-chambers"
        self.court = "in-chambers"

    def set_url(self):
        self.url = "%s/%s.aspx" % (self.url_base, self.court)



# !/usr/bin/python
import argparse
import glob
from lxml import etree
from lxml.etree import _ElementStringResult

__author__ = "mlissner"


def print_xpath_results(query):
    total_result_count = 0
    total_file_count = 0
    equalities = 0
    unique_items = set()
    for f in glob.glob("../examples/*.xml"):
        total_file_count += 1
        indent = ""
        print("\n%s%s:" % (indent, f))
        indent = "    "
        tree = etree.parse(f)
        results = tree.xpath(
            query, namespaces={"m": "http://www.loc.gov/mods/v3"}
        )
        count = 0
        if type(results) in [bool, float]:
            print("%s%s.\t%s" % (indent, count, results))

        elif type(results) == list:
            s = set()
            for result in results:
                if type(result) == _ElementStringResult:
                    print("%s%s.\t%s" % (indent, count, result))
                    s.add("".join(result.split()))
                else:
                    result = etree.tostring(result).strip()
                    print("%s%s.\t%s" % (indent, count, result))
                    s.add("".join(result.split()))
                count += 1
            if len(s) == 1:
                print("\n%sAll items were equal!" % indent)
                equalities += 1
            else:
                print("\n%s%s unique items" % (indent, len(s)))
            unique_items = unique_items.union(s)
        total_result_count += count

    print("\nTotal found: %s" % total_result_count)
    print(
        "All items same in %s/%s sample files" % (equalities, total_file_count)
    )
    print("Unique items: %s/%s" % (len(unique_items), total_result_count))


def main():
    parser = argparse.ArgumentParser(
        description="Apply an XPath query against all the example files, "
        "printing the results."
    )
    parser.add_argument(
        "-q",
        "--query",
        required=True,
        help="XPath query to apply to the example files. Note that "
        "namespaces must be applied to your queries, and that the "
        "mods elements are under the namespace m. So, a query might "
        'look like: "//m:titleInfo"',
    )
    args = parser.parse_args()

    print_xpath_results(args.query)


if __name__ == "__main__":
    main()



from juriscraper.opinions.united_states.state import wash


class Site(wash.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.courtLevel = "C"
        self.pubStatus = "PUB"
        self._set_parameters()

    def _get_case_names(self):
        path = "{base}/td[4]/text()".format(base=self.base)
        return list(self.html.xpath(path))



# Author: Krist Jin
# 2013-08-18: Created.
# 2014-07-17: Updated by mlr to remedy InsanityException.

from lxml import html

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "http://www.state.il.us/court/Opinions/recent_appellate.asp"
        self.base_path = '//table[@class="content"]//table//tr[not(name(..)="thead") and descendant::a][count(td) = 5]'

    def _get_download_urls(self):
        path = self.base_path + "/td[5]//a/@href"
        return list(self.html.xpath(path))

    def _get_case_names(self):
        case_names = []
        for e in self.html.xpath(
            self.base_path + "/td[5]//a[normalize-space(text())]"
        ):
            s = html.tostring(e, method="text", encoding="unicode")
            case_names.append(s)
        return case_names

    def _get_case_dates(self):
        path = self.base_path + "/td[1]/div/text()"
        return [
            convert_date_string(date_string)
            for date_string in self.html.xpath(path)
        ]

    def _get_precedential_statuses(self):
        statuses = []
        for e in self.html.xpath(self.base_path + "/td[3]"):
            s = html.tostring(e, method="text", encoding="unicode")
            if "Rel" in s:
                statuses.append("Unpublished")
            else:
                statuses.append("Published")
        return statuses

    def _get_docket_numbers(self):
        docket_numbers = []
        for e in self.html.xpath(self.base_path + "/td[3]"):
            s = html.tostring(e, method="text", encoding="unicode")
            s = " ".join(s.split())
            s = s.replace("Official Reports", "")
            s = s.replace("NRel", "")
            docket_numbers.append(s)
        return docket_numbers

    def _get_neutral_citations(self):
        neutral_citations = []
        for e in self.html.xpath(self.base_path + "/td[4]"):
            s = html.tostring(e, method="text", encoding="unicode")
            neutral_citations.append(s)
        return neutral_citations



"""Scraper for Missouri
CourtID: mo
Court Short Name: MO
Author: Ben Cassedy
Date created: 04/27/2014
"""

from datetime import date

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url_slug = "Supreme"
        self.url = self.build_url()
        self.cases = []

    def build_url(self):
        url_template = "https://www.courts.mo.gov/page.jsp?id=12086&dist=Opinions %s&date=all&year=%s#all"
        return url_template % (self.url_slug, date.today().year)

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self._extract_cases_from_html(html)
        return html

    def _extract_cases_from_html(self, html):
        for date_block in html.xpath(
            "//div[@id='content']/div/form/table/tr/td"
        ):
            date_string = date_block.xpath("input/@value")
            if date_string:
                for case_block in date_block.xpath("table/tr/td"):
                    links = case_block.xpath("a")
                    first_link_text = links[0].xpath("text()")
                    if (
                        first_link_text
                        and "Orders Pursuant to Rules" in first_link_text[0]
                    ):
                        continue

                    target_link_index = 1 if len(links) > 1 else 0
                    bolded_text = case_block.xpath("b")
                    docket = self.sanitize_docket(
                        bolded_text[0].xpath("text()")[0]
                    )
                    text = case_block.xpath("text()")
                    (
                        judge,
                        disposition,
                    ) = self.parse_judge_disposition_from_text(text)

                    self.cases.append(
                        {
                            "date": convert_date_string(date_string[0]),
                            "docket": docket,
                            "judge": judge,
                            "url": links[target_link_index].xpath("@href")[0],
                            "name": links[target_link_index].xpath("text()")[
                                0
                            ],
                            "disposition": disposition,
                        }
                    )

    @staticmethod
    def sanitize_docket(docket):
        for substring in [":", "and", "_", "Consolidated", "(", ")", ","]:
            docket = docket.replace(substring, " ")
        return ", ".join(docket.split())

    @staticmethod
    def parse_judge_disposition_from_text(text_raw_list):
        text_clean_list = [
            text.strip() for text in text_raw_list if text.strip()
        ]
        return text_clean_list[0], text_clean_list[1]

    def _get_download_urls(self):
        return [case["url"] for case in self.cases]

    def _get_case_names(self):
        return [case["name"] for case in self.cases]

    def _get_case_dates(self):
        return [case["date"] for case in self.cases]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.cases)

    def _get_docket_numbers(self):
        return [case["docket"] for case in self.cases]

    def _get_judges(self):
        return [case["judge"] for case in self.cases]

    def _get_dispositions(self):
        return [case["disposition"] for case in self.cases]



import re

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_code = "S"

    def _download(self, request_dict={}):
        if not self.test_mode_enabled():
            self.url = self.build_url()
        return super(Site, self)._download(request_dict)

    def build_url(self):
        return (
            "http://www.courts.ca.gov/cms/opinions.htm?Courts=%s"
            % self.court_code
        )

    def _get_case_names(self):
        case_names = []
        for cell in self.html.xpath("//table/tr/td[3]"):
            name = cell.text_content()
            date_regex = re.compile(r" \d\d?/\d\d?/\d\d| filed")
            if "P. v. " in date_regex.split(name)[0]:
                case_names.append(
                    date_regex.split(name)[0].replace("P. ", "People ")
                )
            else:
                case_names.append(date_regex.split(name)[0])
        return case_names

    def _get_download_urls(self):
        return [
            t
            for t in self.html.xpath(
                "//table/tr/td[2]/a/@href[contains(.,'PDF')]"
            )
        ]

    def _get_case_dates(self):
        path = "//table/tr/td[1]/text()"
        return [
            convert_date_string(date_string)
            for date_string in self.html.xpath(path)
        ]

    def _get_docket_numbers(self):
        return [t for t in self.html.xpath("//table/tr/td[2]/text()[1]")]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)



"""Scraper for the Michigan Court of Appeals
CourtID: michctapp
Court Short Name: Mich. Ct. App.
Type: Published
Reviewer: mlr
History:
 - 2014-09-19: Created by Jon Andersen
"""

import time
from datetime import date, timedelta

from juriscraper.opinions.united_states.state import mich


class Site(mich.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.today = date.today()
        self.a_while_ago = date.today() - timedelta(days=30)
        self.url = (
            "http://courts.mi.gov/opinions_orders/opinions_orders"
            "/Pages/default.aspx?SearchType=4"
            "&Status_Advanced=coapub&FirstDate_Advanced="
            "{start_month}%2f{start_day}%2f{start_year}"
            "&LastDate_Advanced="
            "{end_month}%2f{end_day}%2f{end_year}".format(
                start_day=self.a_while_ago.day,
                start_month=self.a_while_ago.month,
                start_year=self.a_while_ago.year,
                end_day=self.today.day,
                end_month=self.today.month,
                end_year=self.today.year,
            )
        )
        self.back_scrape_iterable = range(0, 200)
        self.court_id = self.__module__

    def _download_backwards(self, page):
        self.url = (
            "http://courts.mi.gov/opinions_orders/opinions_orders/Pages/default.aspx?SearchType=4&Status_Advanced=coapub&FirstDate_Advanced=1%2f1%2f1900&LastDate_Advanced=9%2f19%2f2014&PageIndex="
            + str(page)
        )
        time.sleep(6)  # This site throttles if faster than 2 hits / 5s.
        self.html = self._download()



from __future__ import print_function
import pprint
import re
import sys

from .reports import BaseReport
from .utils import get_pacer_doc_id_from_doc1_url, reverse_goDLS_function
from ..lib.log_tools import make_default_logger
from ..lib.string_utils import force_unicode

logger = make_default_logger()


class AttachmentPage(BaseReport):
    """An object for querying and parsing the attachment page report. """

    PATH = "doc1/"

    def __init__(self, court_id, pacer_session=None):
        super(AttachmentPage, self).__init__(court_id, pacer_session)
        # Note that parsing bankruptcy attachment pages does not reveal the
        # document number, only the attachment numbers.
        self.is_bankruptcy = self.court_id.endswith("b")

    def query(self, document_number):
        """Query the "attachment page" endpoint and set the results to self.response.

        :param document_number: The internal PACER document ID for the item.
        :return: a request response object
        """
        assert (
            self.session is not None
        ), "session attribute of DocketReport cannot be None."
        # coerce the fourth digit of the document number to 1 to ensure we get
        # the attachment page.
        document_number = document_number[:3] + "0" + document_number[4:]
        url = self.url + document_number
        logger.info(u"Querying the attachment page endpoint at URL: %s", url)
        self.response = self.session.get(url)
        self.parse()

    @property
    def data(self):
        """Get data back from the query for the matching document entry.

        :return: If lookup fails, an empty dict. Else, a dict containing the
        following fields:
            - document_number: The document number we're working with.
            - page_count: The number of pages of the item
            - file_size_str: The size of the file as a str or null if unknown
            - pacer_doc_id: The doc ID for the main document.
            - attachments: A list of attached items with the following fields:
                - attachment_number: The attachment number.
                - description: A description of the item.
                - page_count: The number of pages.
                - file_size_str: The size of the file as a str or null if
                  unknown
                - pacer_doc_id: The document ID for the attachment (a str).

        See the JSON objects in the tests for more examples.
        """
        if self.is_valid is False:
            return {}

        rows = self.tree.xpath("//tr[.//a]")
        if not rows:
            logger.info("No documents found on attachment page.")
            return {}

        first_row = rows.pop(0)
        result = {
            "document_number": self._get_document_number(),
            "page_count": self._get_page_count_from_tr(first_row),
            "file_size_str": self._get_file_size_str_from_tr(first_row),
            "pacer_doc_id": self._get_pacer_doc_id(first_row),
            "pacer_case_id": self._get_pacer_case_id(),
            "pacer_seq_no": self._get_pacer_seq_no_from_tr(first_row),
            "attachments": [],
        }
        if result["document_number"] is None and not self.is_bankruptcy:
            # Abort. Sometimes some attachment pages we receive have a blank
            # area instead of having a proper document number. When going to
            # PACER to reproduce this, things look fine, so I don't know why we
            # get the bad data. In any case, simply give up since we can't do
            # much without a document number.
            return {}
        for row in rows:
            result["attachments"].append(
                {
                    "attachment_number": self._get_attachment_number(row),
                    "description": self._get_description_from_tr(row),
                    "page_count": self._get_page_count_from_tr(row),
                    "file_size_str": self._get_file_size_str_from_tr(row),
                    "pacer_doc_id": self._get_pacer_doc_id(row),
                    # It may not be needed to reparse the seq_no
                    # for each row, but we may as well. So far, it
                    # has always been the same as the main document.
                    "pacer_seq_no": self._get_pacer_seq_no_from_tr(row),
                }
            )

        return result

    def _get_document_number(self):
        """Return the document number for an item.

        In district court attachment pages, this is easy to extract with an
        XPath. In bankruptcy cases, it's simply not there.
        """
        if self.is_bankruptcy:
            return None
        else:
            try:
                path = '//tr[contains(., "Document Number")]//a/text()'
                return int(self.tree.xpath(path)[0].strip())
            except IndexError:
                return None

    def _get_attachment_number(self, row):
        """Return the attachment number for an item.

        In district courts, this can be easily extracted. In bankruptcy courts,
        you must extract it, then subtract 1 from the value since these are
        tallied and include the main document.
        """
        number = int(row.xpath(".//a/text()")[0].strip())
        if self.is_bankruptcy:
            return number - 1
        else:
            return number

    def _get_description_from_tr(self, row):
        """Get the description from the row"""
        if not self.is_bankruptcy:
            index = 2
        else:
            index = 3
        description_text_nodes = row.xpath("./td[%s]//text()" % index)
        if not description_text_nodes:
            # No text in the cell.
            return u""
        else:
            description = description_text_nodes[0].strip()
            return force_unicode(description)

    @staticmethod
    def _get_page_count_from_tr(tr):
        """Take a row from the attachment table and return the page count as an
        int extracted from the cell specified by index.
        """
        pg_cnt_str_nodes = tr.xpath('./td[contains(., "page")]/text()')
        if not pg_cnt_str_nodes:
            # It's a restricted document without page count information.
            return None
        else:
            for pg_cnt_str_node in pg_cnt_str_nodes:
                try:
                    pg_cnt_str = pg_cnt_str_node.strip()
                    return int(pg_cnt_str.split()[0])
                except ValueError:
                    # Happens when the description field contains the
                    # word "page" and gets caught by the xpath. Just
                    # press on.
                    continue

    @staticmethod
    def _get_file_size_str_from_tr(tr):
        """Take a row from the attachment table and return the number of bytes
        as an int.
        """
        cells = tr.xpath("./td")
        last_cell_contents = cells[-1].text_content()
        units = ["kb", "mb"]
        if any(unit in last_cell_contents.lower() for unit in units):
            return last_cell_contents
        return ""

    @staticmethod
    def _get_pacer_doc_id(row):
        """Take in a row from the attachment table and return the pacer_doc_id
        for the item in that row. Return None if the ID cannot be found.
        """
        try:
            url = row.xpath(u".//a")[0]
        except IndexError:
            # Item exists, but cannot download document. Perhaps it's sealed
            # or otherwise unavailable in PACER. This is carried over from the
            # docket report and may not be needed here, but it's a good
            # precaution.
            return None
        else:
            doc1_url = url.xpath("./@href")[0]
            return get_pacer_doc_id_from_doc1_url(doc1_url)

    @staticmethod
    def _get_pacer_seq_no_from_tr(row):
        """Take a row of the attachment page, and return the sequence number
        from the goDLS function.
        """
        try:
            url = row.xpath(u".//a")[0]
        except IndexError:
            # No link in the row. Maybe its sealed.
            pass
        else:
            try:
                onclick = url.xpath("./@onclick")[0]
            except IndexError:
                # No onclick on this row.
                pass
            else:
                if "goDLS" in onclick:
                    go_dls_parts = reverse_goDLS_function(onclick)
                    return go_dls_parts["de_seq_num"]

        # 1. Couldn't find a link in the row: Maybe it's sealed.
        # 2. No onclick on the row.
        # 3. No goDLS in the onclick
        return None

    def _get_pacer_case_id(self):
        """Get the pacer_case_id value by inspecting the HTML

        :returns str: The pacer_case_id value
        """
        # Start by inspecting all the links
        urls = self.tree.xpath("//a")
        for url in urls:
            try:
                onclick = url.xpath("./@onclick")[0]
            except IndexError:
                continue
            else:
                if "goDLS" not in onclick:
                    # Some other onclick we don't care about.
                    continue
                go_dls_parts = reverse_goDLS_function(onclick)
                return go_dls_parts["caseid"]

        # If that fails, try inspecting the input elements
        input_els = self.tree.xpath("//input")
        for input_el in input_els:
            try:
                onclick = input_el.xpath("./@onclick")[0]
            except IndexError:
                continue
            else:
                m = re.search(r"[?&]caseid=(\d+)", onclick, flags=re.I)
                if m:
                    return m.group(1)


def _main():
    if len(sys.argv) != 2:
        print("Usage: python -m juriscraper.pacer.attachment_page filepath")
        print("Please provide a path to an HTML file to parse.")
        sys.exit(1)
    report = AttachmentPage("cand")  # Court ID is only needed for querying.
    filepath = sys.argv[1]
    print("Parsing HTML file at {}".format(filepath))
    with open(filepath, "r") as f:
        text = f.read().decode("utf-8")
    report._parse_text(text)
    pprint.pprint(report.data, indent=2)


if __name__ == "__main__":
    _main()



# coding=utf-8
import copy
import pprint
import re
import sys

import six
from dateutil.tz import gettz
from lxml import etree
from lxml.html import HtmlElement, fromstring, tostring

from .docket_utils import normalize_party_types
from .reports import BaseReport
from .utils import (
    get_pacer_doc_id_from_doc1_url,
    get_pacer_seq_no_from_doc1_anchor,
)
from ..lib.judge_parsers import normalize_judge_string
from ..lib.log_tools import make_default_logger
from ..lib.string_utils import (
    clean_string,
    convert_date_string,
    force_unicode,
    harmonize,
)
from ..lib.utils import clean_court_object, previous_and_next

logger = make_default_logger()

date_regex = r"[—\d\-–/]+"


class BaseDocketReport(object):
    """A small class to hold functions common to the InternetArchive report
    and the PACER DocketReport

    It might be possible to have the InternetArchive report subclass the
    DocketReport, but that brings a lot of cruft along. Better to have this
    little class as a mixin with the common components.
    """

    date_terminated_regex = re.compile(
        r"[tT]erminated:\s+(%s)" % date_regex, flags=re.IGNORECASE
    )
    docket_number_dist_regex = re.compile(
        r"((\d{1,2}:)?\d\d-[a-zA-Z]{1,4}-\d{1,10})"
    )
    docket_number_bankr_regex = re.compile(r"(?:#:\s+)?((\d-)?\d\d-\d*)")
    docket_number_jpml = re.compile(r"(MDL No.\s+\d*)")

    def _clear_caches(self):
        """Clear any caches that are on the object."""
        for attr in self.CACHE_ATTRS:
            setattr(self, "_%s" % attr, None)

    @property
    def data(self):
        """Get all the data back from this endpoint."""
        if self.is_valid is False:
            return {}

        data = self.metadata.copy()
        data[u"parties"] = self.parties
        data[u"docket_entries"] = self.docket_entries
        return data

    @staticmethod
    def _normalize_see_above_attorneys(parties):
        """PACER frequently has "See above" for the contact info of an
        attorney.

        Normalize these values.
        """
        atty_cache = {}
        for party in parties:
            for atty in party.get(u"attorneys", []):
                if not atty[u"contact"]:
                    continue

                if re.search(r"see\s+above", atty[u"contact"], re.I):
                    try:
                        atty_info = atty_cache[atty[u"name"]]
                    except KeyError:
                        # Unable to find the atty in the cache, therefore, we
                        # don't know their contact info.
                        atty[u"contact"] = u""
                    else:
                        # Found the atty in the cache. Use the info.
                        atty[u"contact"] = atty_info
                else:
                    # We have atty info. Save it.
                    atty_cache[atty[u"name"]] = atty[u"contact"]
        return parties

    def _get_value(self, regex, query_strings, cast_to_date=False):
        """Find the matching value for a regex.

        Iterate over a list of values and return group(1) for the first that
        matches regex. If none matches, return the empty string.

        If cast_to_date is True, convert the string to a date object.
        """
        if isinstance(query_strings, six.string_types):
            query_strings = [query_strings]

        for v in query_strings:
            m = regex.search(v)
            if m:
                if cast_to_date:
                    return convert_date_string(m.group(1))
                hit = m.group(1)
                if "date filed" not in hit.lower():
                    # Safety check. Sometimes a match is made against the
                    # merged text string, including its headers. This is wrong.
                    return hit

        if cast_to_date:
            return None
        else:
            return ""

    @staticmethod
    def _xpath_text_0(node, xpath):
        """Get the first text element from a node or return ''

        This is annoyingly hard with normal xpath.
        """
        try:
            return node.xpath("%s/text()" % xpath)[0]
        except IndexError:
            return ""

    def _get_str_from_tree(self, path):
        try:
            s = self.tree.xpath("%s/text()" % path)[0].strip()
        except IndexError:
            return ""  # Return an empty string. Don't return None.
        else:
            return s

    def get_datetime_from_tree(self, path, cast_to_date=False):
        """Parse a datetime from the XML located at node.

        If cast_to_date is true, the datetime object will be converted to a
        date. Else, will return a datetime object in parsed TZ if possible.
        Failing that, it will assume UTC.
        """
        try:
            s = self.tree.xpath("%s/text()" % path)[0].strip()
        except IndexError:
            return None
        else:
            try:
                d = convert_date_string(s, datetime=True)
            except ValueError:
                logger.debug("Couldn't parse date: %s" % s)
                return None
            else:
                # Set it to UTC.
                d = d.replace(tzinfo=d.tzinfo or gettz("UTC"))
                if cast_to_date is True:
                    return d.date()
                return d

    @staticmethod
    def _br_split(element):
        """Split the text of an element on the BRs.

        :param element: Any HTML element
        :return: A list of text nodes from that element split on BRs.
        """
        sep = u"FLP_SEPARATOR"
        html_text = tostring(element, encoding="unicode")
        html_text = re.sub(r"<br/?>", sep, html_text, flags=re.I)
        element = fromstring(html_text)
        text = force_unicode(" ".join(s for s in element.xpath(".//text()")))
        return [s.strip() for s in text.split(sep) if s]

    BR_REGEX = r"(?i)<br\s*/?>"

    @staticmethod
    def redelimit_p(target_element, delimiter_re):
        """Redelimit the children of the target element with <p> tags.

        Insert a <p> tag immediately after the target tag,
        and then replace the delimeter_re with <p> tags.
        Note that <p> is special because the lxml parser knows it
        it self-closing, so this would not work with arbitrary
        tags.

        Use this to turn:
          <foo>a<br>b<br>c</foo>
        Into the more easily iterable:
          <foo>
            <p>a</p>
            <p>b</p>
            <p>c</p>
          </foo>

        :param target_element: An lxml HtmlElement that will be redelimited
        :param delimiter_re: a re pattern matching the tag to replace, e.g.
        BR_REGEX aka r'(?i)<br\s*/?>'
        for a <br> (with optional space and optional /)
        :returns: The redelimited HtmlElement.
        """
        html_text = tostring(target_element, encoding="unicode")
        html_text = re.sub(r"(?i)^(<[^>]*>)", r"\1<p>", html_text)
        html_text = re.sub(delimiter_re, r"<p>", html_text)
        return fromstring(html_text)

    @staticmethod
    def _get_label_value_pair(node, require_colon, field_mappings):
        """Get the field name and value for a node with a tailed value

        :param node: The node with a tailing value.
        :param require_colon: Whether to check for a colon at the end of a
        field name. If True, labels without a colon are ignored.
        :param field_mappings: A dict mapping PACER labels to Juriscraper ones.
        :return a dict with a k-v mapping between a label and its value.
        """
        label = node.text_content().strip()
        if require_colon and not label.endswith(":"):
            return {}
        label = (
            label.strip()
            .lower()
            .replace(" ", "_")
            .replace("(", "")
            .replace(")", "")
            .rstrip(":")
        )
        if six.PY2:
            label = label.decode("utf-8")
        label = field_mappings.get(label, label)

        value = node.tail.strip()
        # Sometimes the colon is in the tail instead of in the label.
        value = value.lstrip(":").strip()
        if label.startswith("date_"):
            # Known date field. Parse it.
            if value:
                data = {label: convert_date_string(value)}
            else:
                data = {label: None}
        else:
            data = {label: force_unicode(value)}
        return data


class DocketReport(BaseDocketReport, BaseReport):
    case_name_str = r"(?:Case\s+title:\s+)?(.*\bvs?\.?\s.*)"
    case_name_regex = re.compile(case_name_str)
    case_name_i_regex = re.compile(case_name_str, flags=re.IGNORECASE)
    case_title_regex = re.compile(
        r"(?:Case\s+title:\s+)(.*)", flags=re.IGNORECASE
    )
    in_re_regex = re.compile(r"(\bIN\s+RE:\s+.*)", flags=re.IGNORECASE)
    in_the_matter_regex = re.compile(
        r"(\bIn\s+the\s+matter\s+.*)", flags=re.IGNORECASE
    )
    case_name_regexes = [
        case_name_regex,
        case_name_i_regex,
        case_title_regex,
        in_re_regex,
        in_the_matter_regex,
    ]
    date_filed_regex = re.compile(r"Date [fF]iled:\s+(%s)" % date_regex)
    date_converted_regex = re.compile(
        r"Date [Cc]onverted:\s+(%s)" % date_regex
    )
    # Be careful this does not match "Joint debtor discharged" field.
    date_discharged_regex = re.compile(
        r"(?:Date|Debtor)\s+[Dd]ischarged:\s+(%s)" % date_regex
    )
    assigned_to_regex = r"Assigned to:\s+(.*)"
    referred_to_regex = r"Referred to:\s+(.*)"
    cause_regex = re.compile(r"Cause:\s+(.*)")
    nos_regex = re.compile(r"Nature of Suit:\s+(.*)")
    jury_demand_regex = re.compile(r"Jury Demand:\s+(.*)")
    jurisdiction_regex = re.compile(r"Jurisdiction:\s+(.*)")
    mdl_status_regex = re.compile(r"MDL Status:\s+(.*)")
    demand_regex = re.compile(r"^Demand:\s+(.*)")
    offense_regex = re.compile(
        r"highest\s+offense.*(?P<status>opening|terminated)", flags=re.I
    )
    counts_regex = re.compile(
        r"(?P<status>pending|terminated)\s+counts", flags=re.I
    )
    complaints_regex = re.compile(r"(?P<status>complaints)", flags=re.I)

    PATH = "cgi-bin/DktRpt.pl"

    CACHE_ATTRS = [
        "metadata",
        "parties",
        "docket_entries",
        "is_adversary_proceeding",
    ]

    ERROR_STRINGS = BaseReport.ERROR_STRINGS + [
        "The report may take a long time to run because this case has many "
        "docket entries",
        "The page ID does not exist. Please enter a valid page ID number. ",
        "There are no documents in this case.",
        "Incomplete request. Please try your query again by choosing the "
        "Query or Reports option",
        "To accept charges shown below, click on the 'View Report' button",
        "Unable to create PDF file.",
        "This case was administratively closed",
        "The start date must be less than or equal to the end date",
        "The starting document number must be less than or equal to the "
        "ending document number",
        "Case not found.",
        "Either you do not have permission to view the document, or the "
        "document does not exist in the case.",
        "Format: text",
        "Server timeout waiting for the HTTP request from the client.",
        "The case type was.*but it must be",
        "This case is in the process of being opened, please check back later "
        "for additional information.",
        "Submission already made, please wait for response from server",
    ]

    def __init__(self, court_id, pacer_session=None):
        super(DocketReport, self).__init__(court_id, pacer_session)
        # Initialize the empty cache properties.
        self._clear_caches()
        self._metadata = None
        self._parties = None
        self._docket_entries = None

        if self.court_id.endswith("b"):
            self.is_bankruptcy = True
        else:
            self.is_bankruptcy = False

    def parse(self):
        """Parse the item, but be sure to clear the cache before you do so.

        This ensures that if the DocketReport is used to parse multiple items,
        the cache is cleared in between.
        """
        self._clear_caches()
        super(DocketReport, self).parse()

    @property
    def metadata(self):
        if self._metadata is not None:
            return self._metadata

        self._set_metadata_values()
        data = {
            u"court_id": self.court_id,
            u"docket_number": self._get_docket_number(),
            u"case_name": self._get_case_name(),
            u"date_filed": self._get_value(
                self.date_filed_regex, self.metadata_values, cast_to_date=True
            ),
            u"date_terminated": self._get_value(
                self.date_terminated_regex,
                self.metadata_values,
                cast_to_date=True,
            ),
            u"date_converted": self._get_value(
                self.date_converted_regex,
                self.metadata_values,
                cast_to_date=True,
            ),
            u"date_discharged": self._get_value(
                self.date_discharged_regex,
                self.metadata_values,
                cast_to_date=True,
            ),
            u"assigned_to_str": self._get_judge(self.assigned_to_regex),
            u"referred_to_str": self._get_judge(self.referred_to_regex),
            u"cause": self._get_value(self.cause_regex, self.metadata_values),
            u"nature_of_suit": self._get_nature_of_suit(),
            u"jury_demand": self._get_value(
                self.jury_demand_regex, self.metadata_values
            ),
            u"demand": self._get_value(
                self.demand_regex, self.metadata_values
            ),
            u"jurisdiction": self._get_value(
                self.jurisdiction_regex, self.metadata_values
            ),
            u"mdl_status": self._get_value(
                self.mdl_status_regex, self.metadata_values
            ),
        }
        data = clean_court_object(data)
        self._metadata = data
        return data

    @property
    def parties(self):
        """Get the party info from the HTML or return it if it's cached.

        The data here will look like this:

            parties = [{
                'name': 'NATIONAL VETERANS LEGAL SERVICES PROGRAM',
                'type': 'Plaintiff',
                'date_terminated': '2018-03-12',
                'extra_info': ("1600 K Street, NW\n"
                               "Washington, DC 20006"),
                'attorneys': [{
                    'name': 'William H. Narwold',
                    'contact': ("1 Corporate Center\n",
                                "20 Church Street\n",
                                "17th Floor\n",
                                "Hartford, CT 06103\n",
                                "860-882-1676\n",
                                "Fax: 860-882-1682\n",
                                "Email: bnarwold@motleyrice.com"),
                    'roles': ['LEAD ATTORNEY',
                              'PRO HAC VICE',
                              'ATTORNEY TO BE NOTICED'],
                }, {
                    ...more attorneys here...
                }],
                'criminal_data': {
                    ...See self._add_criminal_data_to_parties()...
                },
            }, {
                ...more parties (and their attorneys) here...
            }]
        """
        if self._parties is not None:
            return self._parties

        # All sibling rows to the rows that identify this as a party table.
        # We focus on the first td, because sometimes the third td in the
        # document table has bold/underline/italic text.
        path = (
            "//tr["
            # Bankruptcy
            "    ./td[1]//i/b/text() or "
            # Regular district
            "    ./td[1]//b/u/text() or "
            # Adversary proceedings
            '    ./td[1]//b/text()[contains(., "-----")]'
            "]/../tr"
        )
        party_rows = self.tree.xpath(path)

        parties = []
        party = {}
        for prev, row, nxt in previous_and_next(party_rows):
            cells = row.xpath(".//td")
            should_continue = self._test_for_early_continues(row, cells, nxt)
            if should_continue:
                continue

            party, should_continue = self._get_party_type(row, cells, party)
            if should_continue:
                continue

            name_path = (
                u".//b[not(./parent::i)][not(./u)]"
                u'[not(contains(., "------"))]'
            )
            is_party_name_cell = len(cells[0].xpath(name_path)) > 0
            prev_has_disposition = (
                prev is not None and "Disposition" in prev.text_content()
            )
            if is_party_name_cell and not prev_has_disposition:
                element = cells[0].xpath(name_path)[0]
                party[u"name"] = force_unicode(element.text_content().strip())
                party[u"extra_info"] = u"\n".join(
                    force_unicode(s.strip())
                    for s in cells[0].xpath(u".//text()[not(./parent::b)]")
                    if s.strip()
                )
                party[u"date_terminated"] = self._get_value(
                    self.date_terminated_regex,
                    party[u"extra_info"],
                    cast_to_date=True,
                )

            if len(cells) == 3 and party != {}:
                party[u"attorneys"] = self._get_attorneys(cells[2])

            if party != {} and party.get(u"type") is None:
                # Ensure that every record has a type value of some kind.
                party[u"type"] = u"Unknown"

            if party not in parties and party != {}:
                # Sometimes there are dups in the docket. Avoid them.
                parties.append(party)

            if self.is_adversary_proceeding:
                # In adversary proceedings, there are multiple rows under one
                # party type header. Nuke the bulk of the party dict, except
                # for the type so that it's ready for the next iteration.
                party = {u"type": party[u"type"]}
            else:
                party = {}

        parties = self._normalize_see_above_attorneys(parties)

        # Do a second pass to get the criminal data if any and attach it to the
        # correct party.
        if not self.is_bankruptcy:
            self._add_criminal_data_to_parties(parties, party_rows)

        self._parties = parties
        return parties

    @staticmethod
    def _test_for_early_continues(row, cells, nxt):
        """Check for opportunities to skip the current row.

        :param row: The tr that we're parsing.
        :param cells: A list of tds within the row.
        :param nxt: The next row after the current one, so we can look ahead in
        it.
        :returns bool: True if there's an opportunity to continue to the next
        iteration in the current loop; False if not.
        """
        if len(cells) == 0:
            # Empty row. Press on.
            return True
        row_text = force_unicode(row.text_content()).strip().lower()
        if not row_text or row_text == "v.":
            # Empty or nearly empty row. Press on.
            return True

        # Handling for criminal cases that put non-party info in the party html
        # table (see cand, 3:09-cr-00418).
        if cells[0].xpath(".//b/u"):
            if nxt is None or not nxt.xpath(".//b"):
                # If a header is followed by a cell that lacks bold text,
                # punt the header row.
                return True
            if len(cells) == 3 and cells[2].text_content() == "Disposition":
                # This row contains count/offense information. Skip it until
                # criminal data is done in a second pass through the rows.
                return True
        elif not row.xpath(".//b"):
            # If a row has no bold text, we can punt it. The bold normally
            # signifies the party's name, or the party type. This ignores
            # the metadata under these sections.
            return True
        return False

    def _get_party_type(self, row, cells, party):
        """Get the party type info and return it as a dict.

        :param row: The tr we're currently processing.
        :param cells: The array of cells in the row.
        :param party: The party dict as it existed before this iteration. If
        no matches happen in this code, we just pass this through.
        :returns A tuple of the party dict with just the type parameter
        completed and a boolean indicating whether the calling function should
        continue its loop.
        """
        row_text = force_unicode(row.text_content()).strip().lower()
        if len(cells) == 1 and cells[0].xpath(".//b/u"):
            # Regular docket - party type value.
            s = force_unicode(cells[0].text_content())
            return {u"type": normalize_party_types(s)}, True
        elif "------" in row_text:
            # Adversary proceeding
            s = force_unicode(cells[0].text_content().strip())
            if len(cells) == 1:
                s = re.sub(u"----*", "", s)
                return {u"type": normalize_party_types(s)}, True
            elif len(cells) == 3:
                # Some courts have malformed HTML that requires extra work.
                return {u"type": re.split(u"----*", s)[0]}, False
        elif all(
            [self.is_bankruptcy, len(cells) == 3, cells[0].xpath(".//i/b")]
        ):
            # Bankruptcy - party type value.
            s = force_unicode(cells[0].xpath(u".//i")[0].text_content())
            return {u"type": normalize_party_types(s)}, False
        elif len(cells) == 3 and u"service list" in row_text:
            # Special case to handle the service list.
            return {u"type": u"Service List"}, False
        else:
            return party, False

    def _add_criminal_data_to_parties(self, parties, party_rows):
        """Iterate over the party rows, identify criminal data, and add it.

        Final criminal data will look like:

        'criminal_data': {
            'highest_offense_level_opening': 'None',
            'highest_offense_level_terminated': 'Felony',
            'counts': [{
                'name': 'Attempted money laundering',
                'disposition': '',
                'status': 'pending',
            }, {
                'name': 'Theft of public property',
                'disposition': 'Dismissed on deft's motion',
                'status': 'terminated',
            }],
            'complaints': [{
                'name': '18 USC 1956',
                'disposition': '',
            }],
        },

        :param parties: The already-populated party dicts
        :param party_rows: The trs with party/criminal data
        :return: None
        """
        # Because criminal data spans multiple trs, the way we do this is by
        # keeping track of which party we're currently working on. Then, when
        # we get useful criminal data, we add it to that party.
        empty_criminal_data = {
            u"counts": [],
            u"complaints": [],
            u"highest_offense_level_opening": "",
            u"highest_offense_level_terminated": "",
        }
        section_info = {
            "current_section": None,
            "header_info": None,
            "changed": False,
        }
        current_party_i = -1
        criminal_data = copy.deepcopy(empty_criminal_data)
        for prev, row, nxt in previous_and_next(party_rows):
            cells = row.xpath(".//td")
            if len(cells) == 0:
                # Empty row. Press on.
                continue
            row_text = force_unicode(row.text_content()).strip().lower()
            if not row_text or row_text == "v.":
                # Empty or nearly empty row. Press on.
                continue

            is_represented_by_row = len(
                cells
            ) == 3 and "represented by" in clean_string(
                cells[1].text_content()
            )
            # A row representing a party without representation. ID it by
            # looking for bold underlined text in this row followed by bold
            # text in the next row.
            is_special_party_type_row = all(
                [
                    len(cells) == 1,
                    cells[0].xpath(".//b/u"),
                    nxt is not None
                    and nxt.xpath(".//b")
                    and "represented by"
                    not in clean_string(nxt.text_content()),
                ]
            )
            if any([is_represented_by_row, is_special_party_type_row]):
                # If we hit a "represented by" row or a row that looks like a
                # party type row, we know we've moved to the next party.
                # Increment the party index, so we know to whom we should
                # associate the criminal data.
                if criminal_data != empty_criminal_data:
                    criminal_data = clean_court_object(criminal_data)
                    parties[current_party_i][u"criminal_data"] = criminal_data
                current_party_i += 1
                # Reset section info and criminal data.
                criminal_data = copy.deepcopy(empty_criminal_data)
                self._values_to_none(section_info)
                continue

            section_info = self._get_current_section(section_info, cells)
            if section_info["changed"]:
                continue

            if section_info["current_section"] == "highest_offense":
                offense_level = cells[0].text_content()
                criminal_data[section_info["header_info"]] = offense_level
                continue

            if section_info["current_section"] == "counts":
                try:
                    disposition = cells[2].text_content()
                except IndexError:
                    # Sometimes, if there's no disposition, the cell itself is
                    # missing.
                    disposition = ""
                count_name = cells[0].text_content().strip()
                if count_name == u"None":
                    # No counts here. (Happens with terminated ones a lot.)
                    continue
                criminal_data[u"counts"].append(
                    {
                        u"name": count_name,
                        u"disposition": disposition,
                        u"status": section_info["header_info"],
                    }
                )
                continue

            if section_info["current_section"] == "complaints":
                try:
                    disposition = cells[2].text_content()
                except IndexError:
                    # No disposition info.
                    disposition = ""
                complaint_name = cells[0].text_content()
                if complaint_name == u"None":
                    # No counts here. (Happens with terminated ones a lot.)
                    continue
                criminal_data[u"complaints"].append(
                    {u"name": complaint_name, u"disposition": disposition,}
                )

    @staticmethod
    def _values_to_none(d):
        """Set the values for a dictionary all to None"""
        for k in d.keys():
            d[k] = None

    def _get_current_section(self, current_section, cells):
        """Get the current section for the row.

        If the row contains a section header, then we return the new section
        value. If not, then we return the current_section value (without
        changing it).

        :param current_section: The current section when this function is
        started. This might just get passed through.
        :param cells: The cells (tds) in the current row as a list
        :returns dict with keys:
          'current_section': the current section name
          'header_info': any extra info that might be needed by the caller that
                         we gather from the section header itself
          'changed': whether the section has changed, a boolean,
        """
        cell_0_text = clean_string(cells[0].text_content())
        offense_m = self.offense_regex.search(cell_0_text)
        if offense_m:
            return {
                "current_section": u"highest_offense",
                "header_info": u"highest_offense_level_%s"
                % offense_m.group("status").lower(),
                "changed": True,
            }

        counts_m = self.counts_regex.search(cell_0_text)
        if counts_m:
            return {
                "current_section": u"counts",
                "header_info": counts_m.group("status").lower(),
                "changed": True,
            }

        complaint_m = self.complaints_regex.search(cell_0_text)
        if complaint_m:
            return {
                "current_section": u"complaints",
                "header_info": None,
                "changed": True,
            }

        if cells[0].xpath("./b/u"):
            # it's a header we don't recognize like "Plaintiff". We can't hard
            # code them all.
            return {
                "current_section": "unknown",
                "header_info": None,
                "changed": True,
            }

        current_section["changed"] = False
        return current_section

    @staticmethod
    def _get_attorneys(cell):
        """Get the attorney information from an HTML tr node.

        Input will look like:

            <td width="40%" valign="top">
                <b>Allen                Durham          Arnold         </b>
                <br>Arendall &amp; Associates
                <br>2018 Morris Avenue
                <br>Suite 300
                <br>Birmingham              , AL 35203
                <br>205-252-1550
                <br>Fax: 205-252-1556
                <br>Email: ada@arendalllaw.com
                <br><i>LEAD ATTORNEY</i>
                <br><i>ATTORNEY TO BE NOTICED</i><br><br>

                <b>David                Randall         Arendall        </b>
                <br>Arendall &amp; Associates
                <br>2018 Morris Avenue, Third Floor
                <br>Birmingham              , AL 35203
                <br>205-252-1550
                <br>Fax: 205-252-1556
                <br>Email: dra@arendalllaw.com
                <br><i>LEAD ATTORNEY</i>
                <br><i>ATTORNEY TO BE NOTICED</i><br><br>
            </td>

        Output:

            [{
                'name': 'William H. Narwold',
                'contact': ("1 Corporate Center\n",
                            "20 Church Street\n",
                            "17th Floor\n",
                            "Hartford, CT 06103\n",
                            "860-882-1676\n",
                            "Fax: 860-882-1682\n",
                            "Email: bnarwold@motleyrice.com"),
                'roles': ['LEAD ATTORNEY',
                          'PRO HAC VICE',
                          'ATTORNEY TO BE NOTICED'],
            }, {
                ...more attorneys here...
            }]
        """
        attorneys = []
        for atty_node in cell.xpath(".//b"):
            name_parts = force_unicode(
                atty_node.text_content().strip()
            ).split()
            attorney = {
                u"name": u" ".join(name_parts),
                u"roles": [],
                u"contact": u"",
            }
            path = u"./following-sibling::* | ./following-sibling::text()"
            for prev, node, nxt in previous_and_next(atty_node.xpath(path)):
                # noinspection PyProtectedMember
                if isinstance(
                    node,
                    (etree._ElementStringResult, etree._ElementUnicodeResult),
                ):
                    clean_atty = u"%s\n" % " ".join(
                        n.strip() for n in node.split()
                    )
                    if clean_atty.strip():
                        attorney[u"contact"] += clean_atty
                else:
                    if node.tag == u"i":
                        role = force_unicode(node.text_content().strip())
                        if not any(
                            [
                                role.lower().startswith(u"bar status"),
                                role.lower().startswith(u"designation"),
                            ]
                        ):
                            attorney[u"roles"].append(role)

                nxt_is_b_tag = isinstance(nxt, HtmlElement) and nxt.tag == "b"
                if nxt is None or nxt_is_b_tag:
                    # No more data for this attorney.
                    attorneys.append(attorney)
                    break

        return attorneys

    @property
    def docket_entries(self):
        if self._docket_entries is not None:
            return self._docket_entries

        # There can be multiple docket entry tables on a single docket page.
        # See https://github.com/freelawproject/courtlistener/issues/762. ∴ we
        # need to identify the first table, and all following tables. The
        # following tables lack column headers, so we have to use the
        # preceding-sibling tables to make sure it's right.
        docket_header = './/text()[contains(., "Docket Text")]'
        bankr_multi_doc = (
            'not(.//text()[contains(., "Total file size of '
            'selected documents")])'
        )
        footer_multi_doc = 'not(.//text()[contains(., "Footer format:")])'
        docket_entry_rows = self.tree.xpath(
            "//table"
            "[preceding-sibling::table[{dh}] or {dh}]"
            "[{b_multi_doc}]"
            "[{footer_multi_doc}]"
            "/tbody/tr".format(
                dh=docket_header,
                b_multi_doc=bankr_multi_doc,
                footer_multi_doc=footer_multi_doc,
            )
        )[
            1:
        ]  # Skip the first row.

        docket_entries = []
        for row in docket_entry_rows:
            de = {}
            cells = row.xpath(u"./td[not(./input)]")
            if len(cells) == 4:
                # In some instances, the document entry table has an extra
                # column. See almb, 92-04963
                del cells[1]

            date_filed_str = force_unicode(cells[0].text_content())
            if not date_filed_str:
                # Some older dockets have missing dates. Press on.
                continue
            de[u"date_filed"] = convert_date_string(date_filed_str)
            de[u"document_number"] = self._get_document_number(cells[1])
            results = self._get_pacer_doc_id_and_seq_no(
                cells[1], de[u"document_number"]
            )
            de[u"pacer_doc_id"], de[u"pacer_seq_no"] = results[0], results[1]
            de[u"description"] = self._get_description(cells)

            number = de[u"document_number"]
            if number is not None and not number.isdigit():
                # Some courts use the word "doc" instead of a docket number. We
                # skip these for now.
                continue
            docket_entries.append(de)

        docket_entries = clean_court_object(docket_entries)
        self._docket_entries = docket_entries
        return docket_entries

    @property
    def is_adversary_proceeding(self):
        if self._is_adversary_proceeding is not None:
            return self._is_adversary_proceeding

        adversary_proceeding = False
        path = u'//*[text()[contains(., "Adversary Proceeding")]]'
        if self.tree.xpath(path):
            adversary_proceeding = True

        self._is_adversary_proceeding = adversary_proceeding
        return adversary_proceeding

    def query(
        self,
        pacer_case_id,
        date_range_type="Filed",
        date_start=None,
        date_end=None,
        doc_num_start="",
        doc_num_end="",
        show_parties_and_counsel=False,
        show_terminated_parties=False,
        show_list_of_member_cases=False,
        include_pdf_headers=True,
        show_multiple_docs=False,
        output_format="html",
        order_by="date",
    ):
        """Query the docket report and return the results.

        :param pacer_case_id: The internal PACER case ID for a case.
        :param date_range_type: Whether the date range refers to the date items
        were entered into PACER or the date they were filed.
        :param date_start: The start date for the date range (as a date object)
        :param date_end: The end date for the date range (as a date object)
        :param doc_num_start: A range of documents can be requested. This is
        the lower bound of their ID numbers.
        :param doc_num_end: The upper bound of the requested documents.
        :param show_parties_and_counsel: Whether to show the parties and
        counsel in a case (note this adds expense).
        :param show_terminated_parties: Whether to show terminated parties in a
        case (note this adds expense).
        :param show_list_of_member_cases: Whether to show a list of member
        cases (note, this adds expense).
        :param include_pdf_headers: Whether the PDFs should have headers
        containing their metadata.
        :param show_multiple_docs: Show multiple docs at one time.
        :param output_format: Whether to get back the results as a PDF or as
        HTML.
        :param order_by: The ordering desired for the results.
        :return: None. Instead sets self.response attribute and runs
        self.parse()
        """
        # Set up and sanity tests
        assert (
            self.session is not None
        ), "session attribute of DocketReport cannot be None."
        assert bool(pacer_case_id), (
            "pacer_case_id must be truthy, not '%s'" % pacer_case_id
        )

        if date_range_type not in [u"Filed", u"Entered"]:
            raise ValueError(u"Invalid value for 'date_range_type' parameter.")
        if output_format not in [u"html", u"pdf"]:
            raise ValueError(u"Invalid value for 'output_format' parameter.")
        if order_by == u"date":
            order_by = u"oldest date first"
        elif order_by == u"-date":
            order_by = u"most recent date first"
        elif order_by == u"document_number":
            order_by = u"document number"
        else:
            raise ValueError(u"Invalid value for 'order_by' parameter.")

        if show_terminated_parties and not show_parties_and_counsel:
            raise ValueError(
                u"Cannot show terminated parties if parties and "
                u"counsel are not also requested."
            )

        query_params = {
            u"all_case_ids": pacer_case_id,
            u"sort1": order_by,
            u"date_range_type": date_range_type,
            u"output_format": output_format,
            # Any value works in this parameter, but it cannot be blank.
            # Normally this would have a value like '3:12-cv-3879', but that's
            # not even necessary.
            u"case_num": u" "
            # These fields seem to be unnecessary/unused.
            # 'view_comb_doc_text': '',
            # 'PreResetField': '',
            # 'PreResetFields': '',
        }
        if date_start:
            query_params[u"date_from"] = date_start.strftime(u"%m/%d/%Y")
        else:
            # If it's a big docket and you don't filter it in some form you get
            # an intermediate page that says, paraphrasing: "Do you really want
            # to pull that whole, big, docket?" However, if we always make sure
            # to have this field populated, we don't see that page. ∴, always
            # set this value. See #210.
            query_params[u"date_from"] = "1/1/1960"
        if date_end:
            query_params[u"date_to"] = date_end.strftime(u"%m/%d/%Y")
        if doc_num_start:
            query_params[u"documents_numbered_from_"] = str(int(doc_num_start))
        if doc_num_end:
            query_params[u"documents_numbered_to_"] = str(int(doc_num_end))
        if show_parties_and_counsel is True:
            query_params[u"list_of_parties_and_counsel"] = u"on"
        if show_terminated_parties is True:
            query_params[u"terminated_parties"] = u"on"
        if show_list_of_member_cases is True:
            query_params[u"list_of_member_cases"] = u"on"
        if include_pdf_headers is True:
            query_params[u"pdf_header"] = u"1"
        if show_multiple_docs is True:
            query_params[u"view_multi_docs"] = u"on"

        logger.info(
            u"Querying docket report for case ID '%s' with params %s"
            % (pacer_case_id, query_params)
        )

        self.response = self.session.post(
            self.url + "?1-L_1_0-1", data=query_params
        )
        self.parse()

    def _set_metadata_values(self):
        # The first ancestor table of the table cell containing "date filed"
        table = self.tree.xpath(
            # Match any td containing Date [fF]iled
            u'//td[.//text()[contains(translate(., "f", "F"), "Date Filed:")]]'
            # And find its highest ancestor table that lacks a center tag.
            u"/ancestor::table[not(.//center)][last()]"
        )[0]
        cells = table.xpath(u".//td")
        # Convert the <br> separated content into text strings, treating as
        # much as possible as HTML.
        values = []
        for cell in cells:
            clean_texts = [clean_string(s) for s in self._br_split(cell)]
            values.extend(clean_texts)
        values.append(" ".join(values))
        self.metadata_values = values

    @staticmethod
    def _get_pacer_doc_id_and_seq_no(cell, document_number):
        if not document_number:
            return None, None
        else:
            # We find the first link having the document number as text.
            # This is needed because txnb combines the second and third
            # column in their docket report.
            anchors = cell.xpath(u".//a")
            if len(anchors) == 0:
                # Docket entry exists, but cannot download document (it's
                # sealed, a minute entry, or otherwise unavailable in PACER).
                return None, None
            for anchor in anchors:
                if anchor.text_content().strip() == document_number:
                    doc1_url = anchor.xpath("./@href")[0]
                    pacer_doc_id = get_pacer_doc_id_from_doc1_url(doc1_url)
                    pacer_seq_no = get_pacer_seq_no_from_doc1_anchor(anchor)
                    return pacer_doc_id, pacer_seq_no

        # In case none of our URLs can be parsed.
        return None, None

    def _get_document_number(self, cell):
        """Get the document number.

        Some jurisdictions have the number as, "13 (5 pgs)" so some processing
        is needed. See flsb, 09-02199-JKO.
        """
        words = [
            word for phrase in self._br_split(cell) for word in phrase.split()
        ]
        if words:
            first_word = re.sub(u"[\s\u00A0]", "", words[0])
            if self.court_id == u"txnb":
                # txnb merges the second and third columns, so if the first
                # word is a number, return it. Otherwise, assume doc number
                # isn't listed for the item.
                if first_word.isdigit():
                    return first_word
            else:
                return first_word
        return None

    def _get_description(self, cells):
        if self.court_id != u"txnb":
            return force_unicode(cells[2].text_content())

        s = force_unicode(cells[1].text_content())
        # In txnb the second and third columns of the docket entries are
        # combined. The field can have one of four formats. Attempt the most
        # detailed first, then work our way down to just giving up and
        # capturing it all.
        ws = u"[\s\u00A0]"  # Whitespace including nbsp
        regexes = [
            # 2 (23 pgs; 4 docs) Blab blah (happens when attachments exist and
            # page numbers are on)
            u"^{ws}*\d+{ws}+\(\d+{ws}+pgs;{ws}\d+{ws}docs\){ws}+(.*)",
            # 2 (23 pgs) Blab blah (happens when pg nums are on)
            u"^{ws}*\d+{ws}+\(\d+{ws}+pgs\){ws}+(.*)",
            # 2 Blab blah (happens when page nums are off)
            u"^{ws}*\d+{ws}+(.*)",
            # Blab blah (happens when a doc is sealed; can't be downloaded)
            u"^{ws}*(.*)",
        ]
        for regex in regexes:
            try:
                desc = re.search(regex.format(ws=ws), s).group(1)
                break
            except AttributeError:
                continue
        # OK to ignore error below b/c last regex will always match.
        # noinspection PyUnboundLocalVariable
        return desc

    def _get_case_name(self):
        if self.is_bankruptcy:
            # Check if there is somebody described as a debtor
            try:
                return [
                    p
                    for p in self.parties
                    if p[u"type"] == u"Debtor"
                    or p[u"type"] == u"Debtor In Possession"
                ][0][u"name"]
            except IndexError:
                pass

            # This is probably a sub docket to a larger case. Use that title.
            try:
                path = u'//i[contains(., "Lead BK Title")]/following::text()'
                case_name = self.tree.xpath(path)[0].strip()
            except IndexError:
                case_name = u"Unknown Case Title"

            if self.is_adversary_proceeding:
                case_name += u" - Adversary Proceeding"
        else:
            # Skip the last value, it's a concat of all previous values and
            # isn't needed for case name matching.
            case_name = None
            v = self.metadata_values[0]
            for regex in self.case_name_regexes:
                m = regex.search(v)
                if m:
                    case_name = m.group(1)
                    break

            if case_name is None:
                case_name = u"Unknown Case Title"

        return clean_string(harmonize(case_name))

    def _get_docket_number(self):
        if self.is_bankruptcy:
            docket_number_path = "//font"
            # Uses both b/c sometimes the bankr. cases have a dist-style docket
            # number.
            regexes = [
                self.docket_number_dist_regex,
                self.docket_number_bankr_regex,
            ]
        else:
            docket_number_path = "//h3"
            regexes = [self.docket_number_dist_regex, self.docket_number_jpml]
        nodes = self.tree.xpath(docket_number_path)
        string_nodes = [s.text_content() for s in nodes]
        for regex in regexes:
            for s in string_nodes:
                match = regex.search(s)
                if match:
                    return match.group(1)

    def _get_nature_of_suit(self):
        if self.is_adversary_proceeding:
            # Add the next table too, if it contains the nature of suit.
            path = u'//table[contains(., "Nature[s] of")]//tr'
            rows = self.tree.xpath(path)
            nos = []
            for row in rows:
                cell_texts = [
                    force_unicode(s.strip())
                    for s in row.xpath(u"./td[position() > 2]/text()")
                    if s.strip()
                ]
                if len(cell_texts) > 1:
                    nos.append(" ".join(cell_texts))
            return "; ".join(nos) or ""
        else:
            return self._get_value(self.nos_regex, self.metadata_values)

    def _get_judge(self, regex):
        judge_str = self._get_value(re.compile(regex), self.metadata_values)
        if judge_str:
            return normalize_judge_string(judge_str)[0]
        else:
            # No luck getting it in the metadata_values attribute. Broaden
            # the search to look in the entire docket HTML.
            path = '//*[re:match(text(), "%s")]' % regex
            try:
                judge_str = self.tree.re_xpath(path)[0].text_content()
            except IndexError:
                return ""
            judge_str = judge_str.split("to:")[1]
            return normalize_judge_string(judge_str)[0]


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python -m juriscraper.pacer.docket_report filepath")
        print("Please provide a path to an HTML file to parse.")
        sys.exit(1)
    report = DocketReport("cand")  # Court ID is only needed for querying.
    filepath = sys.argv[1]
    print("Parsing HTML file at %s" % filepath)
    with open(filepath, "r") as f:
        text = f.read().decode("utf-8")
    report._parse_text(text)
    pprint.pprint(report.data, indent=2)



from juriscraper.OpinionSite import OpinionSite
from datetime import date
from datetime import datetime
from datetime import timedelta


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = "http://pacer.ca4.uscourts.gov/cgi-bin/opinions.pl"
        self.court_id = self.__module__
        td = date.today()
        self.parameters = {
            "CASENUM": "",
            "FROMDATE": (td - timedelta(days=20)).strftime("%m-%d-%Y"),
            "TITLE": "",
            "TODATE": td.strftime("%m-%d-%Y"),
        }
        self.method = "POST"

    def _get_case_names(self):
        if self.only_get_unpublished:
            # No precedential cases (or else dups)
            path = "//tr/td[4][../td[1]/a/text()[contains(., 'U')]]/text()"
        else:
            path = "//tr/td[4]/text()"
        names = []
        for s in self.html.xpath(path):
            if s.strip():
                names.append(s)
        return names

    def _get_download_urls(self):
        if self.only_get_unpublished:
            path = "//tr/td[1][../td[1]/a/text()[contains(., 'U')]]/a/@href"
        else:
            path = "//tr/td[1]/a/@href"
        return list(self.html.xpath(path))

    def _get_case_dates(self):
        if self.only_get_unpublished:
            path = "//tr/td[3][../td[1]/a/text()[contains(., 'U')]]/text()"
        else:
            path = "//tr/td[3]/text()"
        return [
            datetime.strptime(date_string.strip(), "%Y/%m/%d").date()
            for date_string in self.html.xpath(path)
        ]

    def _get_docket_numbers(self):
        if self.only_get_unpublished:
            path = "//tr/td[2][../td[1]/a/text()[contains(., 'U')]]//text()"
        else:
            path = "//tr/td[2]//text()"
        docket_numbers = []
        for s in self.html.xpath(path):
            if s.strip():
                docket_numbers.append(s)
        return docket_numbers

    def _get_precedential_statuses(self):
        statuses = []
        # using download link, we can get the statuses
        for download_url in self.download_urls:
            file_name = download_url.split("/")[-1]
            if "u" in file_name.lower():
                statuses.append("Unpublished")
            else:
                statuses.append("Published")
        return statuses

    def _download_backwards(self, dt):
        self.end_date = dt + timedelta(days=6)
        self.resource_org_end_date = date(2007, 7, 31)
        # We only get unpublished docs when we're in a period of time during which we have resource.org docs.
        self.only_get_unpublished = self.end_date < self.resource_org_end_date
        self.parameters["FROMDATE"] = dt.strftime("%m-%d-%Y")
        self.parameters["TODATE"] = self.end_date.strftime("%m-%d-%Y")
        self.html = self._download()



import codecs
import os
import sys
import unittest

from setuptools import setup, find_packages
from setuptools.command.install import install

VERSION = "1.27.1"
AUTHOR = "Free Law Project"
EMAIL = "info@free.law"
HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


class TestNetwork(install):
    """Run network test only"""

    description = "run isolated tests that hit the network"

    def run(self):
        loader = unittest.defaultTestLoader
        runner = unittest.TextTestRunner(verbosity=2)
        tests = loader.discover("./tests/network")
        runner.run(tests)


class VerifyVersion(install):
    """Custom command to verify that the git tag matches our version"""

    description = "verify that the git tag matches our version"

    def run(self):
        tag = os.getenv("CIRCLE_TAG")

        if tag is None:
            sys.exit(
                "The 'verify' option is only available in tagged CircleCI container"
            )

        if tag != VERSION:
            message = (
                "Git tag: {0} does not match the version of this app: {1}"
            )
            sys.exit(message.format(tag, VERSION))


setup(
    name="juriscraper",
    description="An API to scrape American court websites for metadata.",
    license="BSD",
    url="https://github.com/freelawproject/juriscraper",
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    maintainer=AUTHOR,
    maintainer_email=EMAIL,
    keywords=["scraping", "legal", "pacer"],
    long_description=read("README.rst"),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["tests*"]),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=[
        "six",
        "argparse",
        "cchardet>=1.1.2",
        "certifi>=2017.4.17",
        "chardet<3.1.0,>=3.0.2",
        "cssselect",
        "feedparser",
        "geonamescache==0.20",
        "html5lib",
        "lxml==3.5.0",
        "python-dateutil==2.5.0",
        "requests==2.20.0",
        "selenium==2.53.6",
        "tldextract",
    ],
    tests_require=["jsondate3-aware", "mock", "vcrpy"],
    include_package_data=True,
    test_suite="tests.test_local",
    cmdclass={"verify": VerifyVersion, "testnetwork": TestNetwork,},
)



"""
Author: Michael Lissner
Date created: 2013-04-05
Revised by Taliah Mirmalek on 2014-02-07
Scraper for the Supreme Court of Arizona
CourtID: ariz
Court Short Name: Ariz.
"""

import time
from datetime import date

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import titlecase


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "http://www.azcourts.gov/opinions/SearchOpinionsMemoDecs.aspx?court=999"

    def _get_download_urls(self):
        path = '//a[contains(@id , "hypCaseNum")]/@href'
        return [t for t in self.html.xpath(path)]

    def _get_case_names(self):
        path = '//span[contains(@id , "lblTitle")]//text()'
        return [titlecase(t.upper()) for t in self.html.xpath(path)]

    def _get_case_dates(self):
        path = '//span[contains(@id , "FilingDate")]//text()'
        return [
            date.fromtimestamp(
                time.mktime(time.strptime(date_string, "%m/%d/%Y"))
            )
            for date_string in self.html.xpath(path)
        ]

    def _get_precedential_statuses(self):
        statuses = []
        path = '//*[contains(@id, "DecType")]/text()'
        for s in self.html.xpath(path):
            if "OPINION" in s:
                statuses.append("Published")
            elif "MEMORANDUM" in s:
                statuses.append("Unpublished")
            else:
                statuses.append("Unknown")
        return statuses

    def _get_docket_numbers(self):
        path = '//a[contains(@id , "hypCaseNum")]//text()'
        return [t for t in self.html.xpath(path)]



"""Oral Argument Audio Scraper for Eighth Circuit Court of Appeals
CourtID: ca8
Court Short Name: 8th Cir.
Author: Brian W. Carver
Date created: 2014-06-21
History:
 - 2014-07-22: download_url fixed by mlr
"""

from datetime import datetime

from juriscraper.OralArgumentSite import OralArgumentSite
from juriscraper.lib.string_utils import clean_if_py3


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "http://media-oa.ca8.uscourts.gov/circ8rss.xml"

    def _download(self, request_dict={}):
        """Go through the items and filter out ones that aren't complete.
        """
        self.items = []
        html_tree = super(Site, self)._download(request_dict=request_dict)
        for item in html_tree.xpath("//item"):
            case_name = item.xpath("./title/text()")[0].split(":", 1)[1]
            if case_name.strip():
                self.items.append(item)

        # Set self.html to None so it can't be misused.
        return None

    def _get_download_urls(self):
        return [item.xpath("./enclosure/@url")[0] for item in self.items]

    def _get_case_names(self):
        case_names = []
        for txt in [item.xpath("./title/text()")[0] for item in self.items]:
            case_name = clean_if_py3(txt).split(": ", 1)[1]
            case_names.append(case_name)
        return case_names

    def _get_case_dates(self):
        case_dates = []
        for txt in [
            item.xpath("./description/text()")[0] for item in self.items
        ]:
            # I can't see it, but there's apparently whitespace or a newline
            # at the end of these dates that has to be removed or we error out.
            case_date = clean_if_py3(txt).split("about ", 1)[1].strip()
            case_dates.append(datetime.strptime(case_date, "%m/%d/%Y").date())
        return case_dates

    def _get_docket_numbers(self):
        docket_numbers = []
        for txt in [item.xpath("./title/text()")[0] for item in self.items]:
            docket_number = clean_if_py3(txt).split(": ", 1)[0]
            docket_numbers.append(docket_number)
        return docket_numbers



import re

from ..lib.string_utils import titlecase


def normalize_party_types(t):
    """Normalize various party types to as few as possible."""
    t = t.lower().strip()

    # Numerical types
    t = re.sub(r"defendant\s+\(\d+\)", r"defendant", t)
    t = re.sub(r"debtor\s+\d+", "debtor", t)

    # Assorted other
    t = re.sub(r"(thirdparty|3rd pty|3rd party)", r"third party", t)
    t = re.sub(r"(fourthparty|4th pty|4th party)", r"fourth party", t)
    t = re.sub(r"counter-(defendant|claimaint)", r"counter \1", t)
    t = re.sub(r"\bus\b", "u.s.", t)
    t = re.sub(r"u\. s\.", "u.s.", t)
    t = re.sub(r"united states", "u.s.", t)
    t = re.sub(r"jointadmin", "jointly administered", t)
    t = re.sub(r"consolidated-debtor", "consolidated debtor", t)
    t = re.sub(r"plaintiff-? consolidated", "consolidated plaintiff", t)
    t = re.sub(r"defendant-? consolidated", "consolidated defendant", t)
    t = re.sub(r"intervenor-plaintiff", "intervenor plaintiff", t)
    t = re.sub(r"intervenor pla\b", "intervenor plaintiff", t)
    t = re.sub(r"intervenor dft\b", "intervenor defendant", t)

    return titlecase(t)



"""Scraper for the Supreme Court of Michigan
CourtID: mich
Court Short Name: Mich.
Contact: sitefeedback@courts.mi.gov
History:
 - 2014-09-21: Updated by Jon Andersen to handle some fields being missing
 - 2014-08-05: Updated to have a dynamic URL, an oversight during check in.
"""

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import titlecase, clean_if_py3
import time
from datetime import date, timedelta


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.today = date.today()
        self.a_while_ago = date.today() - timedelta(days=30)
        self.url = (
            "http://courts.mi.gov/opinions_orders/opinions_orders/"
            "Pages/default.aspx?SearchType=4"
            "&Status_Advanced=sct&FirstDate_Advanced="
            "{start_month}%2f{start_day}%2f{start_year}"
            "&LastDate_Advanced="
            "{end_month}%2f{end_day}%2f{end_year}".format(
                start_day=self.a_while_ago.day,
                start_month=self.a_while_ago.month,
                start_year=self.a_while_ago.year,
                end_day=self.today.day,
                end_month=self.today.month,
                end_year=self.today.year,
            )
        )
        self.back_scrape_iterable = range(0, 868)
        self.court_id = self.__module__

    def _get_case_dates(self):
        dates = []
        for txt in self.html.xpath('//li[@class="releaseDate"]/text()'):
            # Release Date: 2/22/2013 --> 2/22/2013
            txt = clean_if_py3(txt).strip().split(" ")[2]
            dates.append(
                date.fromtimestamp(
                    time.mktime(time.strptime(txt.strip(), "%m/%d/%Y"))
                )
            )
        return dates

    def _get_docket_numbers(self):
        docket_numbers = []
        for s in self.html.xpath(
            '//li[@class="casenumber" and not(ancestor::ul[@class="result-header"])]'
        ):
            docket_numbers.append(s.text)
        return docket_numbers

    def _get_case_names(self):
        case_names = []
        for case_name in self.html.xpath(
            '//li[@class="title1" and not(ancestor::ul[@class="result-header"])]/a/text()'
        ):
            case_name = titlecase(case_name)
            if "People of Mi " in case_name:
                case_name = case_name.replace(
                    "People of Mi ", "People of Michigan "
                )
            case_names.append(case_name)
        return case_names

    def _get_download_urls(self):
        return [
            s
            for s in self.html.xpath(
                '//li[@class="title1" and not(ancestor::ul[@class="result-header"])]/a/@href'
            )
        ]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)

    def _get_lower_courts(self):
        lower_courts = []
        for el in self.html.xpath(
            "//ul[contains(@class, 'odd') or contains(@class, 'even')]"
        ):
            try:
                s = el.xpath(
                    '//li[@class="casedetailsleft"]'
                    '//li[@class="lowerCourt"]/text()'
                )
                lower_courts.append(
                    titlecase(clean_if_py3(s[0]).strip().split(" ", 2)[2])
                )
            except IndexError:
                lower_courts.append("")
        return lower_courts

    def _get_dispositions(self):
        disps = []
        for el in self.html.xpath(
            "//ul[contains(@class, 'odd') or contains(@class, 'even')]"
        ):
            try:
                s = el.xpath('//li[@class="caseNature"]/text()')
                disps.append(clean_if_py3(s[0]).strip().split(" ", 2)[2])
            except IndexError:
                disps.append("")
        return disps

    def _get_lower_court_numbers(self):
        nums = []
        for el in self.html.xpath(
            "//ul[contains(@class, 'odd') or contains(@class, 'even')]"
        ):
            try:
                s = el.xpath(
                    '//li[@class = "casedetailsright"]'
                    '//li[@class = "lowerCourt"]/text()'
                )
                nums.append(clean_if_py3(s[0]).strip().split("No. ")[1])
            except IndexError:
                nums.append("")
        return nums

    def _download_backwards(self, page):
        self.url = (
            "http://courts.mi.gov/opinions_orders/opinions_orders/Pages/default.aspx?SearchType=4&Status_Advanced=sct&FirstDate_Advanced=3%2f1%2f2013&LastDate_Advanced=8%2f8%2f2014&PageIndex="
            + str(page)
        )
        time.sleep(6)  # This site throttles if faster than 2 hits / 5s.
        self.html = self._download()



from juriscraper.OpinionSite import OpinionSite
import time
from datetime import date
from datetime import timedelta
from juriscraper.lib.string_utils import clean_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        begin_date = date.strftime(date.today() - timedelta(15), "%m/%d/%Y")
        end_date = date.strftime(date.today(), "%m/%d/%Y")
        self.url = (
            "http://wicourts.gov/supreme/scopin.jsp?"
            "begin_date=%s&end_date=%s&SortBy=date"
        ) % (begin_date, end_date)
        self.court_id = self.__module__

    def _get_case_names(self):
        return [t for t in self.html.xpath("//form/table[1]//tr/td[3]/text()")]

    def _get_download_urls(self):
        return [
            href
            for href in self.html.xpath(
                '//form/table[1]/tr//td[4]/a[contains(.//text(), "PDF")]/@href'
            )
        ]

    def _get_case_dates(self):
        dates = []
        for date_string in self.html.xpath("//form/table[1]/tr//td[1]/text()"):
            dates.append(
                date.fromtimestamp(
                    time.mktime(
                        time.strptime(clean_string(date_string), "%b %d, %Y")
                    )
                )
            )
        return dates

    def _get_docket_numbers(self):
        return [e for e in self.html.xpath("//form/table[1]/tr//td[2]/text()")]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)



"""
History:
 - 2014-08-05: Adapted scraper to have year-based URLs.
"""

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = (
            "https://www.courts.oregon.gov/publications/sc/Pages/default.aspx"
        )
        self.cases = []
        self.status = "Published"

    def _process_html(self):
        for header in self.html.xpath("//h4//a/parent::h4"):
            date_string = header.text_content().strip()
            if not date_string:
                continue
            ul = header.xpath("./following-sibling::ul")[0]
            for item in ul.xpath(".//li"):
                # Ensure two links are present (skip Petitions for Review rows)
                # see or_example_2.html
                anchors = item.xpath(".//a")
                if not (len(anchors) > 1):
                    continue
                text = item.text_content().strip()
                url = anchors[0].xpath("./@href")[0]
                docket = anchors[1].text_content().strip()
                name = text.split(")", 1)[-1]
                self.cases.append(
                    {
                        "date": date_string,
                        "name": name,
                        "docket": docket,
                        "url": url,
                    }
                )



from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import titlecase
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = "http://www.cafc.uscourts.gov/opinions-orders?field_origin_value=All&field_report_type_value=All"
        self.back_scrape_iterable = range(1, 700)
        self.court_id = self.__module__

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self._extract_cases_from_html(html)
        return html

    def _extract_cases_from_html(self, html):
        """Build list of data dictionaries, one dictionary per case (table row)."""
        self.cases = []

        for row in html.xpath("//table/tbody/tr"):
            date, docket, url, name, status = False, False, False, False, False

            date = convert_date_string(row.xpath("td[1]/span/text()")[0])
            docket = row.xpath("td[2]/text()")[0].strip()

            url_raw = row.xpath("td[4]/a/@href")
            if url_raw:
                url = url_raw[0]

            name_raw = row.xpath("td[4]/a/text()")
            if name_raw:
                name = titlecase(name_raw[0].split("[")[0].strip())

            status_raw = row.xpath("td[5]/text()")
            if status_raw:
                status_raw = status_raw[0].strip().lower()
                if "nonprecedential" in status_raw.lower():
                    status = "Unpublished"
                elif "precedential" in status_raw.lower():
                    status = "Published"
                else:
                    status = "Unknown"

            if date and docket and url and name and status:
                self.cases.append(
                    {
                        "date": date,
                        "docket": docket,
                        "url": url,
                        "name": name,
                        "status": status,
                    }
                )

    def _get_case_names(self):
        return [case["name"] for case in self.cases]

    def _get_download_urls(self):
        return list(self.html.xpath("//table//tr/td[4]/a/@href"))

    def _get_case_dates(self):
        return [case["date"] for case in self.cases]

    def _get_docket_numbers(self):
        return [case["docket"] for case in self.cases]

    def _get_precedential_statuses(self):
        return [case["status"] for case in self.cases]

    def _download_backwards(self, n):
        self.url = "http://www.cafc.uscourts.gov/opinions-orders?page={}".format(
            n
        )

        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200



import ssl
import time
from datetime import date

from dateutil.rrule import MONTHLY, rrule
from juriscraper.OpinionSite import OpinionSite

from juriscraper.lib.network_utils import SSLAdapter


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = "https://www.cadc.uscourts.gov/internet/opinions.nsf/OpinionsByMonday?OpenView&StartKey=20151020150928&Count=2&scode=1"
        self.court_id = self.__module__
        self.back_scrape_iterable = [
            i.date()
            for i in rrule(
                MONTHLY, dtstart=date(1997, 9, 1), until=date(2015, 1, 1),
            )
        ]

    def _get_adapter_instance(self):
        """Unfortunately this court doesn't support modern crypto, so you have
        to manually downgrade the crypto it uses.

        See: http://stackoverflow.com/questions/14102416/
        """
        return SSLAdapter(ssl_version=ssl.PROTOCOL_TLSv1)

    def _get_case_names(self):
        return [
            e
            for e in self.html.xpath(
                "//div[@class='row-entry'][position() mod 2 = 1]/span[2]/text()"
            )
        ]

    def _get_download_urls(self):
        return [
            e
            for e in self.html.xpath(
                "//div[@class='row-entry'][position() mod 2 = 1]/span[1]/a/@href"
            )
        ]

    def _get_case_dates(self):
        dates = []
        for date_string in self.html.xpath(
            "//div[@class='row-entry'][position() mod 2 = 0]/span[2]/text()"
        ):
            dates.append(
                date.fromtimestamp(
                    time.mktime(time.strptime(date_string, "%m/%d/%Y"))
                )
            )
        return dates

    def _get_docket_numbers(self):
        return [
            e
            for e in self.html.xpath(
                "//div[@class='row-entry'][position() mod 2 = 1]/span[1]/a/text()"
            )
        ]

    def _get_precedential_statuses(self):
        return ["Published" for _ in range(0, len(self.case_names))]

    def _download_backwards(self, d):

        self.url = "https://www.cadc.uscourts.gov/internet/opinions.nsf/OpinionsByRDate?OpenView&count=100&SKey={}".format(
            d.strftime("%Y%m")
        )

        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200



# -*- coding: utf-8 -*-

import os
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from juriscraper.AbstractSite import phantomjs_executable_path
from juriscraper.lib.cookie_utils import normalize_cookies
from juriscraper.lib.html_utils import clean_html
from juriscraper.lib.html_utils import get_html_parsed_text
from juriscraper.lib.html_utils import fix_links_but_keep_anchors


class WebDriven:
    def __init__(self, *args, **kwargs):
        self.cookies = {}
        self.url = False
        self.uses_selenium = True
        self.wait = False
        self.webdriver = False

    def __del__(self):
        if self.webdriver:
            self.webdriver.quit()

    def get_page(self):
        text = clean_html(self.webdriver.page_source)
        html = get_html_parsed_text(text)
        html.rewrite_links(fix_links_but_keep_anchors, base_href=self.url)
        return html

    def initiate_webdriven_session(self):
        if not self.url:
            raise Exception("self.url not set")
        self.webdriver = webdriver.PhantomJS(
            executable_path=phantomjs_executable_path,
            service_args=["--ignore-ssl-errors=true", "--ssl-protocol=any"],
            # uncomment line below to see webdriver log
            service_log_path=os.path.devnull,
        )
        self.webdriver.implicitly_wait(30)
        self.webdriver.set_window_size(5000, 3000)
        self.wait = WebDriverWait(self.webdriver, 10)
        self.webdriver.get(self.url)
        self.cookies = normalize_cookies(self.webdriver.get_cookies())

    def wait_for_id(self, id_attr):
        self.wait.until(EC.presence_of_element_located((By.ID, id_attr)))

    """Use this method to snap screenshots during debugging"""

    def take_screenshot(self, name="screenshot.png"):
        self.webdriver.save_screenshot(name)



from juriscraper.opinions.united_states.state import nm


class Site(nm.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = (
            "https://nmonesource.com/nmos/nmca/en/nav_date.do?%s" % self.IFRAME
        )



# -*- coding: utf-8 -*-
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import clean_string

from datetime import date
from lxml import html
import re
import time


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = "http://www.isc.idaho.gov/opinions/cacivil.htm"
        self.court_id = self.__module__

    def tweak_response_object(self):
        """
        This state uses an HTTP header to indicate an encoding of UTF-8,
        but the encoding is actually ISO-8859-1, as indicated in the HTML
        itself. When this page is rendered in Firefox, chardet fails, and
        many unknown characters are displayed to the user. Setting the
        character encoding manually fixes the issue.
        """
        self.request["response"].encoding = "ISO-8859-1"

    def _get_case_names(self):
        case_names = []
        for xelement in self.html.xpath('//li[@class="MsoNormal"]/span/a[1]'):
            # Our text nodes are randomly surrounded by HTML spans, so we start
            # at a high level node, and strip down to just what we want.
            data_html = html.tostring(xelement, encoding=unicode)
            data_string = re.sub(r"<[^>]*?>", "", data_html)
            data_string = " ".join(data_string.split())

            # Several regexes are needed since we have hyphen-separated values
            regexes = [
                u"(.*?)[-–](.*?)–",
                u"(.*?)[-–](.*)-",
                u"(.*?)[-–](.*)$",
            ]
            for regex in regexes:
                try:
                    case_names.append(re.search(regex, data_string).group(2))
                    # Found what we're looking for...
                    break
                except AttributeError:
                    # Try the next regex...
                    continue

        return case_names

    def _get_download_urls(self):
        download_urls = []
        for e in self.html.xpath('//li[@class="MsoNormal"]/span/a[1]/@href'):
            download_urls.append(e)
        return download_urls

    def _get_case_dates(self):
        case_dates = []
        for e in self.html.xpath('//li[@class="MsoNormal"]/span/a[1]'):
            s = html.tostring(e, method="text", encoding="unicode")
            # Cleanup...
            s = s.split(u"-")[0]
            s = s.split(u"–")[0]
            date_formats = ["%B %d, %Y", "%B %d %Y", "%B %d , %Y"]
            for format in date_formats:
                try:
                    case_date = date.fromtimestamp(
                        time.mktime(time.strptime(clean_string(s), format))
                    )
                except ValueError:
                    continue
            case_dates.append(case_date)
        return case_dates

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)



# -*- coding: utf-8 -*-

from juriscraper.OpinionSiteLinear import OpinionSiteLinear
from juriscraper.WebDriven import WebDriven


class OpinionSiteLinearWebDriven(OpinionSiteLinear, WebDriven):
    def __init__(self, *args, **kwargs):
        super(OpinionSiteLinearWebDriven, self).__init__(*args, **kwargs)
        WebDriven.__init__(self, args, kwargs)
